//---------------------------------------------------------------------------
/*
Copyright (c) 2013 Serge L. Ryadkow admin@opcgate.ru http://opcgate.ru 

Permission is hereby granted, free of charge, to any person obtaining a copy of this 
software and associated documentation files (the "Software"), to deal in the Software 
without restriction, including without limitation the rights to use, copy, modify, 
merge, publish, distribute, sublicense, and/or sell copies of the Software, and to 
permit persons to whom the Software is furnished to do so, subject to the following 
conditions:

The above copyright notice and this permission notice shall be included in all copies 
or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, 
INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR 
PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE 
FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR 
OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR 
OTHER DEALINGS IN THE SOFTWARE. 
*/
//---------------------------------------------------------------------------
#include <vcl.h>

#pragma hdrstop

#include "Base64.h"

//---------------------------------------------------------------------------
#pragma package(smart_init)
//---------------------------------------------------------------------------
namespace Base64 {
//---------------------------------------------------------------------------
// Translation Table as described in RFC1113
static const char cb64[]="ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789+/";
//---------------------------------------------------------------------------
// Translation Table to decode
static const char cd64[]="|$$$}rstuvwxyz{$$$$$$$>?@ABCDEFGHIJKLMNOPQRSTUVW$$$$$$XYZ[\\]^_`abcdefghijklmnopq";
//---------------------------------------------------------------------------
// encodeblock
// encode 3 8-bit binary bytes as 4 '6-bit' characters
void EncodeBlock(unsigned char *in, unsigned char *out, int length)
{
  out[0] = cb64[in[0] >> 2];
  out[1] = cb64[((in[0] & 0x03) << 4) | ((in[1] & 0xf0) >> 4)];
  out[2] = (unsigned char)(length > 1 ? cb64[((in[1] & 0x0f) << 2) | ((in[2] & 0xc0) >> 6)] : '=');
  out[3] = (unsigned char)(length > 2 ? cb64[in[2] & 0x3f] : '=');
}
//---------------------------------------------------------------------------
// decodeblock
// decode 4 '6-bit' characters into 3 8-bit binary bytes
void DecodeBlock(unsigned char *in, unsigned char *out)
{
  out[0] = (unsigned char) (in[0] << 2 | in[1] >> 4);
  out[1] = (unsigned char) (in[1] << 4 | in[2] >> 2);
  out[2] = (unsigned char) (((in[2] << 6) & 0xc0) | in[3]);
}
//---------------------------------------------------------------------------
AnsiString Encode(void *src, int length, bool nocrlf)
{
  AnsiString s;
  unsigned char in[3], out[5] = {0, 0, 0, 0, 0};
  for (int i = 0, n = 0; i < length;) {
    int l = 0;
    for (int j = 0; j < 3; ++j) {
      if (i < length) {
        in[j] = reinterpret_cast<unsigned char*>(src)[i++];
        ++l;
      } else
        in[j] = 0;
    }
    EncodeBlock(in, out, l);
    s += reinterpret_cast<char*>(out);
    if ((++n >= 19 || i == length) && !nocrlf) {
      s += "\r\n";
      n = 0;
    }
  }
  return s;
}
//---------------------------------------------------------------------------
int Decode(const AnsiString &src, void *dst, int dstSize)
{
  unsigned char in[4], out[3], v;
  int cnt = 0;
  for (int i = 0, l; i < src.Length();) {
    l = 0;
    for (int j = 0; j < 4 && i < src.Length(); ++j) {
      v = 0;
      while (i < src.Length() && !v) {
        v = reinterpret_cast<unsigned char*>(src.c_str())[i++];
        v = (unsigned char)((v < 43 || v > 122) ? 0 : cd64[v - 43]);
        if (v)
          v = (unsigned char)((v == '$') ? 0 : v - 61);
      }
      if (i < src.Length()) {
        ++l;
        in[j] = v ? (unsigned char)(--v) : 0;
      } else
        in[j] = 0;
    }
    if (l) {
      DecodeBlock(in, out);
      for(int i = 0; i < l - 1; ++i) {
        if (cnt < dstSize)
          reinterpret_cast<unsigned char*>(dst)[cnt] = out[i];
        ++cnt;
      }
    }
  }
  return cnt;
}
//---------------------------------------------------------------------------
} //namespace Base64 {
//---------------------------------------------------------------------------

