//---------------------------------------------------------------------------
/*
Copyright (c) 2013 Serge L. Ryadkow admin@opcgate.ru http://opcgate.ru 

Permission is hereby granted, free of charge, to any person obtaining a copy of this 
software and associated documentation files (the "Software"), to deal in the Software 
without restriction, including without limitation the rights to use, copy, modify, 
merge, publish, distribute, sublicense, and/or sell copies of the Software, and to 
permit persons to whom the Software is furnished to do so, subject to the following 
conditions:

The above copyright notice and this permission notice shall be included in all copies 
or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, 
INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR 
PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE 
FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR 
OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR 
OTHER DEALINGS IN THE SOFTWARE. 
*/
//---------------------------------------------------------------------------

#include <vcl.h>
#pragma hdrstop

#include "Comm.h"
#include "StringUtil.h"

//---------------------------------------------------------------------------

#pragma package(smart_init)
//---------------------------------------------------------------------------

//comm utilite functions
//---------------------------------------------------------------------------
void Normalise(TDCB *dcb)
{
  memset(dcb, 0, sizeof(*dcb));
  dcb->DCBlength = sizeof(*dcb);
  dcb->fParity = true;
}

void Normalise(TCommConfig *cfg)
{
  memset(cfg, 0, sizeof(*cfg));
  cfg->dwSize = sizeof(*cfg);
  cfg->wVersion = 1;
  cfg->dwProviderSubType = PST_UNSPECIFIED;
  cfg->dwProviderOffset = 0;
  cfg->dwProviderSize = 0;
  Normalise(&cfg->dcb);
}

AnsiString CommEventToString(DWORD ev)
{
  AnsiString s = "#";
  s += int(ev);
  s += "   ";
  switch (ev) {
    case EV_BREAK:
      s += "EV_BREAK   A break was detected on input";
      break;
    case EV_CTS:
      s += "EV_CTS   The CTS (clear-to-send) signal changed state";
      break;
    case EV_DSR:
      s += "EV_DSR   The DSR (data-set-ready) signal changed state";
      break;
    case EV_ERR:
      s += "EV_ERR   A line-status error occurred. Line-status errors are CE_FRAME, CE_OVERRUN, and CE_RXPARITY";
      break;
    case EV_RING:
      s += "EV_RING   A ring indicator was detected";
      break;
    case EV_RLSD:
      s += "EV_RLSD   The RLSD (receive-line-signal-detect) signal changed state";
      break;
    case EV_RXCHAR:
      s += "EV_RXCHAR   A character was received and placed in the input buffer";
      break;
    case EV_RXFLAG:
      s += "EV_RXFLAG   The event character was received and placed in the input buffer. The event character is specified in the device's DCB structure, which is applied to a serial port by using the SetCommState function";
      break;
    case EV_TXEMPTY:
      s += "EV_TXEMPTY   The last character in the output buffer was sent";
      break;
    default:
      s += "Unknown COMM event";
  }
  return s;
}

bool CommConfigDialog(
    AnsiString name,	// device name string
    HWND hWnd,	      // handle to window
    TCommConfig *cfg	// pointer to comm. configuration structure
   )
{
  bool ret = true;
  if (!CommConfigDialog(name.c_str(), hWnd, cfg))
    if (GetLastError() == 1223) //Cancel pressed
      ret = false;
    else
      RaiseLastWin32Error();
  return ret;
}

void BuildCommDCB(
    AnsiString def,	// device-control string
    TDCB *dcb	      // pointer to device-control block
   )
{
  Win32Check(BuildCommDCB(def.c_str(), dcb));
}

void BuildCommDCBAndTimeouts(
    AnsiString def,	  // device-control string
    TDCB *dcb,	      // pointer to device-control block
    TCommTimeouts *to //pointer to time-out structure
   )
{
  Win32Check(BuildCommDCBAndTimeouts(def.c_str(), dcb, to));
}

void GetDefaultCommConfig(
    AnsiString name,	// device name string
    TCommConfig *cfg	// pointer to configuration structure
   )
{
  DWORD s1 = sizeof(*cfg);
  DWORD s2 = s1;
  Win32Check(GetDefaultCommConfig(name.c_str(), cfg, &s1));
  if (s1 != s2)
    throw Exception("Invalid structure length");
}

void SetDefaultCommConfig(
    AnsiString name,	// device name string
    TCommConfig *cfg	// pointer to configuration structure
   )
{
  Win32Check(SetDefaultCommConfig(name.c_str(), cfg, sizeof(*cfg)));
}

//---------------------------------------------------------------------------

//class TComm
//---------------------------------------------------------------------------
__fastcall TComm::TComm(const AnsiString &deviceName)
  : THandleStream(reinterpret_cast<int>(::CreateFile(
    ("\\\\.\\" + deviceName).c_str(), GENERIC_READ | GENERIC_WRITE,
    0, 0, OPEN_EXISTING, FILE_ATTRIBUTE_NORMAL, 0)))
{
  ::Win32Check(Handle != reinterpret_cast<int>(INVALID_HANDLE_VALUE));
}
//---------------------------------------------------------------------------
__fastcall TComm::TComm(const AnsiString &deviceName, const AnsiString &cfg)
  : THandleStream(reinterpret_cast<int>(::CreateFile(
    ("\\\\.\\" + deviceName).c_str(), GENERIC_READ | GENERIC_WRITE,
    0, 0, OPEN_EXISTING, FILE_ATTRIBUTE_NORMAL, 0)))
{
  ::Win32Check(Handle != reinterpret_cast<int>(INVALID_HANDLE_VALUE));
  try {
    TDCB dcb;
    Normalise(&dcb);
    BuildCommDCB(cfg, &dcb);
    SetDCB(&dcb);
  }
  catch(...) {
    ::CloseHandle(reinterpret_cast<HANDLE>(Handle));
    throw;
  }
}
//---------------------------------------------------------------------------
__fastcall TComm::~TComm()
{
  ::CloseHandle(reinterpret_cast<HANDLE>(Handle));
}
//---------------------------------------------------------------------------
void __fastcall TComm::SetDCB(TDCB *dcb)
{
  ::Win32Check(::SetCommState(reinterpret_cast<HANDLE>(Handle), dcb));
}
//---------------------------------------------------------------------------
void __fastcall TComm::GetDCB(TDCB *dcb)
{
  ::Win32Check(::GetCommState(reinterpret_cast<HANDLE>(Handle), dcb));
}
//---------------------------------------------------------------------------
void __fastcall TComm::SetConfig(TCommConfig *cfg)
{
  ::Win32Check(::SetCommConfig(reinterpret_cast<HANDLE>(Handle), cfg, sizeof(*cfg)));
}
//---------------------------------------------------------------------------
void __fastcall TComm::GetConfig(TCommConfig *cfg)
{
  unsigned long size = sizeof(*cfg);
  ::Win32Check(::GetCommConfig(reinterpret_cast<HANDLE>(Handle), cfg, &size));
}
//---------------------------------------------------------------------------
void __fastcall TComm::SetMask(DWORD value)
{
  ::Win32Check(::SetCommMask(reinterpret_cast<HANDLE>(Handle), value));
}
//---------------------------------------------------------------------------
DWORD __fastcall TComm::GetMask()
{
  DWORD ret;
  ::Win32Check(::GetCommMask(reinterpret_cast<HANDLE>(Handle), &ret));
  return ret;
}
//---------------------------------------------------------------------------
void __fastcall TComm::Setup(DWORD inQueue, DWORD outQueue)
{
  ::Win32Check(::SetupComm(reinterpret_cast<HANDLE>(Handle), inQueue, outQueue));
}
//---------------------------------------------------------------------------
void __fastcall TComm::SetTimeouts(TCommTimeouts *to)
{
  ::Win32Check(::SetCommTimeouts(reinterpret_cast<HANDLE>(Handle), to));
}
//---------------------------------------------------------------------------
void __fastcall TComm::GetTimeouts(TCommTimeouts *to)
{
  ::Win32Check(::GetCommTimeouts(reinterpret_cast<HANDLE>(Handle), to));
}
//---------------------------------------------------------------------------
DWORD __fastcall TComm::Wait()
{
  DWORD evt = 0;
  ::Win32Check(::WaitCommEvent(reinterpret_cast<HANDLE>(Handle), &evt, 0));
  return evt;
}
//---------------------------------------------------------------------------
void __fastcall TComm::GetStat(COMSTAT *comstat)
{
  DWORD err = 0;
  ::Win32Check(::ClearCommError(reinterpret_cast<HANDLE>(Handle), &err, comstat));
  if (err) {
    AnsiString es = "Communication error #";
    es += int(err);
    throw Exception(es);
  }
}
//---------------------------------------------------------------------------
void __fastcall TComm::Purge(DWORD mask)
{
  ::Win32Check(::PurgeComm(reinterpret_cast<HANDLE>(Handle), mask));
}
//---------------------------------------------------------------------------
//---------------------------------------------------------------------------
bool IsComm(const AnsiString &s)
{
  if (s.SubString(1, 3).UpperCase() != "COM")
    return false;
  int i;
  for (i = 3; i < s.Length(); ++i)
    if (!isdigit(s.c_str()[i]))
      if (s.c_str()[i] != ' ')
        return false;
      else
        break;
  if (i == 3)
    return false;
  int n = s.SubString(4, i - 3).ToInt();
  if (n < 1 || n > 256)
    return false;
  else
    return true;
}
//---------------------------------------------------------------------------
bool IsSocket(const AnsiString &s)
{
  bool ret;
  TStringList *sl = new TStringList;
  try {
    if (SplitString(s, sl, ":") != 2)
      ret = false;
    else
      if (IsIPAddress(sl->Strings[0]))
        if (sl->Strings[1].ToIntDef(-1) >= 0)
          ret = true;
        else
          ret = false;
      else
        ret = false;
  }
  __finally {
    delete sl;
  }
  return ret;
}
//---------------------------------------------------------------------------
AnsiString GetCommString(const AnsiString &s)
{
  if (!IsComm(s))
    throw Exception("Invalid COM device \"" + s + "\"");
  return s.SubString(1, s.Pos(" ")).Trim();
}
//---------------------------------------------------------------------------
AnsiString GetIPAddr(const AnsiString &s)
{
  if (!IsSocket(s))
    throw Exception("Invalid socket address \"" + s + "\"");
  AnsiString ret;
  TStringList *sl = new TStringList;
  try {
    SplitString(s, sl, ":");
    ret = sl->Strings[0];
  }
  __finally {
    delete sl;
  }
  return ret;
}
//---------------------------------------------------------------------------
int GetIPPort(const AnsiString &s)
{
  if (!IsSocket(s))
    throw Exception("Invalid socket address \"" + s + "\"");
  int ret;
  TStringList *sl = new TStringList;
  try {
    SplitString(s, sl, ":");
    ret = sl->Strings[1].ToInt();
  }
  __finally {
    delete sl;
  }
  return ret;
}
//---------------------------------------------------------------------------
//---------------------------------------------------------------------------
bool __fastcall TCommStringParser::GetIsComm()
{
  return ::IsComm(SrcString);
}
//---------------------------------------------------------------------------

