//---------------------------------------------------------------------------
/*
Copyright (c) 2013 Serge L. Ryadkow admin@opcgate.ru http://opcgate.ru 

Permission is hereby granted, free of charge, to any person obtaining a copy of this 
software and associated documentation files (the "Software"), to deal in the Software 
without restriction, including without limitation the rights to use, copy, modify, 
merge, publish, distribute, sublicense, and/or sell copies of the Software, and to 
permit persons to whom the Software is furnished to do so, subject to the following 
conditions:

The above copyright notice and this permission notice shall be included in all copies 
or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, 
INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR 
PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE 
FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR 
OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR 
OTHER DEALINGS IN THE SOFTWARE. 
*/
//---------------------------------------------------------------------------
//---------------------------------------------------------------------------
#include <Vcl.h>

#pragma hdrstop

#include <process.h>

#include "FileVersion.h"
#include "StringUtil.h"

//---------------------------------------------------------------------------
#pragma package(smart_init)
//---------------------------------------------------------------------------
int GetFileVersion(TStrings *sl, const AnsiString &fn)
{
	sl->Clear();
  const AnsiString is[10] = {"CompanyName",
                             "FileDescription",
                             "FileVersion",
                             "InternalName",
                             "LegalCopyright",
                             "LegalTradeMarks",
                             "OriginalFileName",
                             "ProductName",
                             "ProductVersion",
                             "Comments"};
  int l1 = 0;
  unsigned long l = GetFileVersionInfoSize(fn.c_str(), &l);
  if (l) {
    char *b = new char[l];
    try {
      Win32Check(GetFileVersionInfo(fn.c_str(), 0, l, b));
      char *v;
      UINT l;
      if (VerQueryValue(b, "\\VarFileInfo\\Translation", (void**)&v, &l)) {
        l1 = *(short*)v;
        int l2 = *(short*)(v + 2);
        AnsiString s = "\\StringFileInfo\\" + IntToHex(l1, 4) + IntToHex(l2, 4) + "\\";
        for (int i = 0; i < 10; ++i)
          if (VerQueryValue(b, (s + is[i]).c_str(), (void**)&v, &l))
            sl->Add(is[i] + "=" + v);
      }
    }
    __finally {
	    delete[] b;
    }
  }
  return l1;
}
//---------------------------------------------------------------------------
int __fastcall GetFileVersionLo(const AnsiString &fn)
{
  unsigned long l = GetFileVersionInfoSize(fn.c_str(), &l);
  int ver = 0;
  if (l) {
    char *b = new char[l];
    try {
      Win32Check(GetFileVersionInfo(fn.c_str(), 0, l, b));
      VS_FIXEDFILEINFO *v;
      UINT l;
      if (VerQueryValue(b, "\\", (void**)&v, &l))
        ver = v->dwFileVersionLS;
    }
    __finally {
	    delete[] b;
    }
  }
  return ver;
}
//---------------------------------------------------------------------------
int __fastcall GetFileVersionHi(const AnsiString &fn)
{
  unsigned long l = GetFileVersionInfoSize(fn.c_str(), &l);
  int ver = 0;
  if (l) {
    char *b = new char[l];
    try {
      Win32Check(GetFileVersionInfo(fn.c_str(), 0, l, b));
      VS_FIXEDFILEINFO *v;
      UINT l;
      if (VerQueryValue(b, "\\", (void**)&v, &l))
        ver = v->dwFileVersionMS;
    }
    __finally {
	    delete[] b;
    }
  }
  return ver;
}
//---------------------------------------------------------------------------
AnsiString __fastcall GetFileVersionString(const AnsiString &fn)
{
  int hi = GetFileVersionHi(fn);
  int lo = GetFileVersionLo(fn);
  return AnsiString(hi >> 16) + "." +
         AnsiString(hi & 0xffff) + "." +
         AnsiString(lo >> 16) + "." +
         AnsiString(lo & 0xffff);
}
//---------------------------------------------------------------------------
void __fastcall GetFileVersionStrings(TStrings *sl, const AnsiString &fn)
{
  sl->Clear();
  int hi = GetFileVersionHi(fn);
  int lo = GetFileVersionLo(fn);
  sl->Add(hi >> 16);
  sl->Add(hi & 0xffff);
  sl->Add(lo >> 16);
  sl->Add(lo & 0xffff);
}
//---------------------------------------------------------------------------
bool __fastcall IsNewVersion(const AnsiString &oldfn, const AnsiString &newfn)
{
  int o = GetFileVersionHi(oldfn);
  int n = GetFileVersionHi(newfn);
  if (n > o)
    return true;
  else if (n == o)
    if (GetFileVersionLo(oldfn) < GetFileVersionLo(newfn))
      return true;
    else
      return false;
  else
    return false;
}
//---------------------------------------------------------------------------
void __fastcall UpdateVersion(const AnsiString &path, bool prompt, bool bakup)
{
  AnsiString ap = Application->ExeName;
  AnsiString ext = ExtractFileExt(ap).UpperCase();
  if (ext == ".NEW") {
    AnsiString en = ChangeFileExt(ap, ".exe");
    AnsiString bn = ChangeFileExt(ap, ".bak");
    Win32Check(CopyFile(en.c_str(), bn.c_str(), false));
    Win32Check(!execl(bn.c_str(), bn.c_str(), 0));
  } else if (ext == ".BAK") {
    AnsiString nn = ChangeFileExt(ap, ".new");
    AnsiString en = ChangeFileExt(ap, ".exe");
    Win32Check(MoveFileEx(nn.c_str(), en.c_str(), MOVEFILE_REPLACE_EXISTING));
    Win32Check(!execl(en.c_str(), en.c_str(), 0));
  } else {
    if (!bakup) {
      AnsiString bn = ChangeFileExt(ap, ".bak");
      if (FileExists(bn))
        Win32Check(DeleteFile(bn));
    }
    AnsiString un = NormPath(path) + ExtractFileName(ap);
    if (FileExists(un)) {
      if (IsNewVersion(ap, un)) {
        if (prompt) {
          TStringList *ov = new TStringList;
          TStringList *nv = new TStringList;
          AnsiString msg;
          try {
            GetFileVersionStrings(nv, un);
            GetFileVersionStrings(ov, ap);
            msg = "   : " +
                  nv->Strings[0] + "." + nv->Strings[1] +
                  "\n: " + nv->Strings[2] +
                  " : " + nv->Strings[3] +
                  "\n\n : " +
                  ov->Strings[0] + "." + ov->Strings[1] +
                  "\n: " + ov->Strings[2] +
                  " : " + ov->Strings[3] +
                  "\n\n  "
                  "\n   ?";
          }
          __finally {
            delete ov;
            delete nv;
          }
          if (bakup)
            msg += "\n\n    "
                   "\n    \"BAK\"";
          int ret = MessageBox(0, msg.c_str(), " ",
                               MB_YESNO | MB_ICONQUESTION);
        	Win32Check(ret);
          if (ret != IDYES)
            return;
        }
        AnsiString uh = ChangeFileExt(un, ".hlp");
        if (FileExists(uh)) {
          AnsiString hp = ChangeFileExt(ap, ".hlp");
          //delete .cnt and so on
          Win32Check(CopyFile(uh.c_str(), hp.c_str(), false));
        }
        AnsiString nn = ChangeFileExt(ap, ".new");
        Win32Check(CopyFile(un.c_str(), nn.c_str(), false));
        Win32Check(!execl(nn.c_str(), nn.c_str(), 0));
      }
    }
  }
}
//---------------------------------------------------------------------------

