//---------------------------------------------------------------------------
/*
Copyright (c) 2013 Serge L. Ryadkow admin@opcgate.ru http://opcgate.ru 

Permission is hereby granted, free of charge, to any person obtaining a copy of this 
software and associated documentation files (the "Software"), to deal in the Software 
without restriction, including without limitation the rights to use, copy, modify, 
merge, publish, distribute, sublicense, and/or sell copies of the Software, and to 
permit persons to whom the Software is furnished to do so, subject to the following 
conditions:

The above copyright notice and this permission notice shall be included in all copies 
or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, 
INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR 
PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE 
FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR 
OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR 
OTHER DEALINGS IN THE SOFTWARE. 
*/
//---------------------------------------------------------------------------
//---------------------------------------------------------------------------
#ifndef FilterH
#define FilterH
//---------------------------------------------------------------------------
template<class T>
class TFilter {
public:
  virtual T __fastcall Filter(const T &value) = 0;
};
//---------------------------------------------------------------------------
template<class T>
class TFilterBuff {
public:
  void __fastcall Clear();
  bool __fastcall IsFull();
protected:
  __fastcall TFilterBuff(int size);
  __fastcall ~TFilterBuff();
  void __fastcall Add(const T &value);
  int Count;
  T *Buff;
private:
  const int Size;
};
//---------------------------------------------------------------------------
template<class T>
class TAvgFilter : public TFilter<T>, public TFilterBuff<T> {
public:
  __fastcall TAvgFilter(int size);
  virtual T __fastcall Filter(const T &value);
};
//---------------------------------------------------------------------------
template<class T>
class TMedFilter : public TFilter<T>, public TFilterBuff<T> {
public:
  __fastcall TMedFilter(int size);
  __fastcall ~TMedFilter();
  virtual T __fastcall Filter(const T &value);
private:
  T **PBuff;
};
//---------------------------------------------------------------------------
template<class T>
class TMinFilter : public TFilter<T>, public TFilterBuff<T> {
public:
  __fastcall TMinFilter(int size);
  virtual T __fastcall Filter(const T &value);
};
//---------------------------------------------------------------------------
template<class T>
class TMaxFilter : public TFilter<T>, public TFilterBuff<T> {
public:
  __fastcall TMaxFilter(int size);
  virtual T __fastcall Filter(const T &value);
};
//---------------------------------------------------------------------------
//---------------------------------------------------------------------------
/////////////////////////////////////////////////////////////////////////////
//Implementations
/////////////////////////////////////////////////////////////////////////////
//---------------------------------------------------------------------------
//---------------------------------------------------------------------------
//TFilterBuff implementations
//---------------------------------------------------------------------------
template<class T>
void __fastcall TFilterBuff<T>::Clear()
{
  Count = 0;
}
//---------------------------------------------------------------------------
template<class T>
bool __fastcall TFilterBuff<T>::IsFull()
{
  return Count == Size;
}
//---------------------------------------------------------------------------
template<class T>
__fastcall TFilterBuff<T>::TFilterBuff(int size)
	: Size(size),
    Count(0)
{
  if (Size < 2)
  	throw Exception("    \n"
                    "    \n"
                    "(  " + AnsiString(size) + ")");
	Buff = new T[size];
}
//---------------------------------------------------------------------------
template<class T>
__fastcall TFilterBuff<T>::~TFilterBuff()
{
	delete[] Buff;
}
//---------------------------------------------------------------------------
template<class T>
void __fastcall TFilterBuff<T>::Add(const T& value)
{
  memmove(Buff + 1, Buff, (Size - 1) * sizeof(T));
  Buff[0] = value;
  if (Count < Size)
    ++Count;
}
//---------------------------------------------------------------------------
//---------------------------------------------------------------------------
//TAvgFilter implementations
//---------------------------------------------------------------------------
template<class T>
__fastcall TAvgFilter<T>::TAvgFilter(int size)
	:	TFilterBuff<T>(size)
{
}
//---------------------------------------------------------------------------
template<class T>
T __fastcall TAvgFilter<T>::Filter(const T &value)
{
	Add(value);
  T v = 0;
  for (int i = 0; i < Count; ++i)
    v += Buff[i];
  return v / Count;
}
//---------------------------------------------------------------------------
//---------------------------------------------------------------------------
//TMedFilter implementations
//---------------------------------------------------------------------------
template<class T>
__fastcall TMedFilter<T>::TMedFilter(int size)
	:	TFilterBuff<T>(size)
{
	if (!(size & 1))
  	throw Exception("   \n"
                    "   \n"
                    "(  " + AnsiString(size) + ")");
  PBuff = new T*[size];
}
//---------------------------------------------------------------------------
template<class T>
__fastcall TMedFilter<T>::~TMedFilter()
{
	delete[] PBuff;
}
//---------------------------------------------------------------------------
template<class T>
T __fastcall TMedFilter<T>::Filter(const T &value)
{
	Add(value);
  if (Count > 2) {
    int cnt = Count & 1 ? Count : Count - 1;
    for (int i = 0; i < cnt; ++i)
    	PBuff[i] = Buff + i;
    for (int k = cnt / 2; k > 0; k /= 2)
      for (int i = k; i < cnt; ++i)
        for (int j = i - k; j >= 0 && *PBuff[j] > *PBuff[j + k]; j -= k) {
          T *t = PBuff[j];
          PBuff[j] = PBuff[j + k];
          PBuff[j + k] = t;
        }
    return *PBuff[cnt / 2];
  } else
    return Buff[0];
}
//---------------------------------------------------------------------------
//---------------------------------------------------------------------------
//TMinFilter implementations
//---------------------------------------------------------------------------
template<class T>
__fastcall TMinFilter<T>::TMinFilter(int size)
	:	TFilterBuff<T>(size)
{
}
//---------------------------------------------------------------------------
template<class T>
T __fastcall TMinFilter<T>::Filter(const T &value)
{
	Add(value);
  T v = 0;
  for (int i = 0; i < Count; ++i)
    if (!i || Buff[i] < v)
      v = Buff[i];
  return v;
}
//---------------------------------------------------------------------------
//---------------------------------------------------------------------------
//TMaxFilter implementations
//---------------------------------------------------------------------------
template<class T>
__fastcall TMaxFilter<T>::TMaxFilter(int size)
	:	TFilterBuff<T>(size)
{
}
//---------------------------------------------------------------------------
template<class T>
T __fastcall TMaxFilter<T>::Filter(const T &value)
{
	Add(value);
  T v = 0;
  for (int i = 0; i < Count; ++i)
    if (!i || Buff[i] > v)
      v = Buff[i];
  return v;
}
//---------------------------------------------------------------------------
#endif
