//---------------------------------------------------------------------------
/*
Copyright (c) 2013 Serge L. Ryadkow admin@opcgate.ru http://opcgate.ru 

Permission is hereby granted, free of charge, to any person obtaining a copy of this 
software and associated documentation files (the "Software"), to deal in the Software 
without restriction, including without limitation the rights to use, copy, modify, 
merge, publish, distribute, sublicense, and/or sell copies of the Software, and to 
permit persons to whom the Software is furnished to do so, subject to the following 
conditions:

The above copyright notice and this permission notice shall be included in all copies 
or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, 
INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR 
PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE 
FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR 
OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR 
OTHER DEALINGS IN THE SOFTWARE. 
*/
//---------------------------------------------------------------------------
//---------------------------------------------------------------------------
#include <vcl.h>

#pragma hdrstop

#include <DateUtils.hpp>

#include "Ipc.h"
#include "Util.h"
//---------------------------------------------------------------------------
#pragma package(smart_init)
//---------------------------------------------------------------------------
namespace IPC {
//---------------------------------------------------------------------------
THndl::~THndl() {
	if (Hndl)
		Win32Check(CloseHandle(FHndl));
}
//---------------------------------------------------------------------------
TSharedMemory::TSharedMemory(AnsiString name, int size)
	:	FPtr(0) {
 	FHndl = OpenFileMapping(
    FILE_MAP_ALL_ACCESS,	// access mode
    true,	// inherit flag
    name.c_str() 	// pointer to name of file-mapping object
	);
  if (!size)
   	Win32Check(FHndl != 0);
	if (!FHndl) {
	 	FHndl = CreateFileMapping(
  	  (HANDLE)0xFFFFFFFF,	// Hndl to file to map
    	0,	// optional security attributes
	    PAGE_READWRITE,	// protection for mapping object
  	  0,	// high-order 32 bits of object size
    	(DWORD)size,	// low-order 32 bits of object size
	    name.c_str() 	// name of file-mapping object
		);
   	Win32Check(FHndl != 0);
  	FExisting = false;
  } else
  	FExisting = true;
	FPtr = MapViewOfFile(
    FHndl,	// file-mapping object to map into address space
    FILE_MAP_ALL_ACCESS,	// access mode
    0,	// high-order 32 bits of file offset
    0,	// low-order 32 bits of file offset
    0 	// number of bytes to map
	);
  Win32Check(Ptr != 0);
}
//---------------------------------------------------------------------------
TSharedMemory::TSharedMemory(int size)
	:	FPtr(0) {
 	FHndl = CreateFileMapping(
 	  (HANDLE)0xFFFFFFFF,	// Hndl to file to map
   	0,	// optional security attributes
    PAGE_READWRITE,	// protection for mapping object
 	  0,	// high-order 32 bits of object size
   	(DWORD)size,	// low-order 32 bits of object size
    0 	// name of file-mapping object
	);
 	Win32Check(FHndl != 0);
	FPtr = MapViewOfFile(
    FHndl,	// file-mapping object to map into address space
    FILE_MAP_ALL_ACCESS,	// access mode
    0,	// high-order 32 bits of file offset
    0,	// low-order 32 bits of file offset
    0 	// number of bytes to map
	);
  Win32Check(Ptr != 0);
}
//---------------------------------------------------------------------------
TSharedMemory::TSharedMemory(THndl *file, AnsiString name, int size)
	:	FPtr(0) {
 	FHndl = OpenFileMapping(
    FILE_MAP_ALL_ACCESS,	// access mode
    true,	// inherit flag
    name.c_str() 	// pointer to name of file-mapping object
	);
  if (!size)
   	Win32Check(FHndl != 0);
	if (!FHndl) {
	 	FHndl = CreateFileMapping(
  	  file->Hndl,	// Hndl to file to map
    	0,	// optional security attributes
	    PAGE_READWRITE,	// protection for mapping object
  	  0,	// high-order 32 bits of object size
    	(DWORD)size,	// low-order 32 bits of object size
	    name.c_str() 	// name of file-mapping object
		);
   	Win32Check(FHndl != 0);
  }
	FPtr = MapViewOfFile(
    FHndl,	// file-mapping object to map into address space
    FILE_MAP_ALL_ACCESS,	// access mode
    0,	// high-order 32 bits of file offset
    0,	// low-order 32 bits of file offset
    0 	// number of bytes to map
	);
  Win32Check(Ptr != 0);
}
//---------------------------------------------------------------------------
TSharedMemory::TSharedMemory(THndl *file, int size)
	:	FPtr(0) {
 	FHndl = CreateFileMapping(
 	  file->Hndl,	// Hndl to file to map
   	0,	// optional security attributes
    PAGE_READWRITE,	// protection for mapping object
 	  0,	// high-order 32 bits of object size
   	(DWORD)size,	// low-order 32 bits of object size
    0 	// name of file-mapping object
	);
 	Win32Check(FHndl != 0);
	FPtr = MapViewOfFile(
    FHndl,	// file-mapping object to map into address space
    FILE_MAP_ALL_ACCESS,	// access mode
    0,	// high-order 32 bits of file offset
    0,	// low-order 32 bits of file offset
    0 	// number of bytes to map
	);
  Win32Check(Ptr != 0);
}
//---------------------------------------------------------------------------
TSharedMemory::~TSharedMemory() {
	Win32Check(UnmapViewOfFile(Ptr));
}
//---------------------------------------------------------------------------
bool TSyncObj::Wait(int timeout) {
	switch(WaitForSingleObject(
    Hndl,	// Hndl of object to wait for
    timeout 	// time-out interval in milliseconds
  )) {
  	case WAIT_OBJECT_0:
    	return true;
   	case WAIT_FAILED:
    	RaiseLastOSError();
    case WAIT_ABANDONED:
    case WAIT_TIMEOUT:
			return false;
    default:
    	throw(Exception("Unknown TSyncObj::Wait() state"));
  }
}
//---------------------------------------------------------------------------
TSyncObj *
WaitAny(TSyncObj *o1, TSyncObj *o2, int timeout) {
	HANDLE h[2];
  h[0] = o1->Hndl;
  h[1] = o2->Hndl;
	switch(WaitForMultipleObjects(
    2,	// number of Hndls in the object Hndl array
    h,	// pointer to the object-Hndl array
    false,	// wait all flag
    timeout 	// time-out interval in milliseconds
  )) {
  	case WAIT_OBJECT_0 + 0:
    	return o1;
  	case WAIT_OBJECT_0 + 1:
    	return o2;
   	case WAIT_FAILED:
    	RaiseLastOSError();
    case WAIT_ABANDONED_0 + 0:
    case WAIT_ABANDONED_0 + 1:
    case WAIT_TIMEOUT:
			return 0;
    default:
    	throw(Exception("Unknown WaitAny state"));
  }
}
//---------------------------------------------------------------------------
/*
TSyncObj *
WaitAny(TList *l, int timeout) {
	HANDLE *h = new HANDLE[l->Count];
  for (int i = 0; i < l->Count; ++i)
  	h[i] = static_cast<TSyncObj*>(l->Items[i])->Hndl;
  DWORD ret = WaitForMultipleObjects(
  	            l->Count,	// number of Hndls in the object Hndl array
                h,	// pointer to the object-Hndl array
	              false,	// wait all flag
  	            timeout 	// time-out interval in milliseconds
              );
  delete[] h;
  for (int i = 0; i < l->Count; ++i) {
		if (ret == WAIT_OBJECT_0 + i)
    	return static_cast<TSyncObj*>(l->Items[i]);
	  if (ret == WAIT_ABANDONED_0 + i)
			return 0;
  }
	switch(ret) {
 		case WAIT_FAILED:
   		RaiseLastOSError();
    case WAIT_TIMEOUT:
			return 0;
   	default:
   		throw(Exception("Unknown WaitAny state"));
  }
}
//---------------------------------------------------------------------------
TSyncObj *
WaitAny(TList *l, TSyncObj *o, int timeout) {
	HANDLE *h = new HANDLE[l->Count + 1];
  for (int i = 0; i < l->Count; ++i)
  	h[i] = static_cast<TSyncObj*>(l->Items[i])->Hndl;
  h[l->Count] = o->Hndl;
  DWORD ret = WaitForMultipleObjects(
  	            l->Count,	// number of Hndls in the object Hndl array
                h,	// pointer to the object-Hndl array
	              false,	// wait all flag
  	            timeout 	// time-out interval in milliseconds
              );
  delete[] h;
  for (int i = 0; i <= l->Count; ++i) {
		if (ret == WAIT_OBJECT_0 + i)
    	if (i < l->Count)
	    	return static_cast<TSyncObj*>(l->Items[i]);
      else
      	return o;
	  if (ret == WAIT_ABANDONED_0 + i)
			return 0;
  }
	switch(ret) {
 		case WAIT_FAILED:
   		RaiseLastOSError();
    case WAIT_TIMEOUT:
			return 0;
   	default:
   		throw(Exception("Unknown WaitAny state"));
  }
}
*/
//---------------------------------------------------------------------------
TMutex::TMutex(AnsiString name) {
 	FHndl = OpenMutex(
    MUTEX_ALL_ACCESS,	// access mode
    true,	// inherit flag
    name.c_str() 	// pointer to name of mutex object
	);
	if (!FHndl) {
	 	FHndl = CreateMutex(
    	0,	// optional security attributes
	    false,	// flag for initial ownership
	    name.c_str() 	// name of mutex object
		);
   	Win32Check(FHndl != 0);
  }
}
//---------------------------------------------------------------------------
TMutex::TMutex() {
 	FHndl = CreateMutex(
   	0,	// optional security attributes
    false,	// flag for initial ownership
    0 	// name of mutex object
	);
 	Win32Check(FHndl != 0);
}
//---------------------------------------------------------------------------
void TMutex::Release() {
	Win32Check(ReleaseMutex(FHndl));
}
//---------------------------------------------------------------------------
TMutex::TLock::TLock(TMutex *mutex, int timeout)
	: Mutex(mutex) {
 	if (!Mutex->Wait(timeout))
   	throw(Exception("Mutex lock timeout"));
}
//---------------------------------------------------------------------------
TMutex::TLock::~TLock() {
 	Mutex->Release();
}
//---------------------------------------------------------------------------
TEvent::TEvent(AnsiString name) {
 	FHndl = OpenEvent(
    EVENT_ALL_ACCESS,	// access mode
    true,	// inherit flag
    name.c_str() 	// pointer to name of mutex object
	);
	if (!FHndl) {
	 	FHndl = CreateEvent(
    	0,	// optional security attributes
			false,	// flag for manual-reset event
    	false,	// flag for initial state
	    name.c_str() 	// name of mutex object
		);
   	Win32Check(FHndl != 0);
  }
}
//---------------------------------------------------------------------------
TEvent::TEvent() {
 	FHndl = CreateEvent(
   	0,	// optional security attributes
		false,	// flag for manual-reset event
   	false,	// flag for initial state
    0 	// name of mutex object
	);
 	Win32Check(FHndl != 0);
}
//---------------------------------------------------------------------------
void
TEvent::Set() {
	Win32Check(SetEvent(FHndl));
}
//---------------------------------------------------------------------------
void
TEvent::Reset() {
	Win32Check(ResetEvent(FHndl));
}
//---------------------------------------------------------------------------
TChangeNotification::TChangeNotification(AnsiString path, bool subtree, int filter) {
 	FHndl = FindFirstChangeNotification(
    path.c_str(), // pointer to name of directory to watch
    subtree, // flag for monitoring directory or directory tree
    filter 	// filter conditions to watch for
  );
 	Win32Check(FHndl != INVALID_HANDLE_VALUE);
}
//---------------------------------------------------------------------------
TChangeNotification::~TChangeNotification() {
	Win32Check(FindCloseChangeNotification(
    FHndl 	// handle to change notification to close
  ));
  FHndl = 0; //to prevent CloseHandle()
}
//---------------------------------------------------------------------------
bool
TChangeNotification::Wait(int timeout) {
	if (TSyncObj::Wait(timeout)) {
  	Next();
    return true;
  } else
  	return false;
}
//---------------------------------------------------------------------------
void
TChangeNotification::Next() {
	Win32Check(FindNextChangeNotification(
    FHndl 	// handle to change notification to signal
  ));
}
//---------------------------------------------------------------------------
TWaitableTimer::TWaitableTimer(AnsiString name) {
 	FHndl = OpenWaitableTimer(
    TIMER_ALL_ACCESS,	// access mode
    true,	// inherit flag
    name.c_str() 	// pointer to name of timer object
	);
	if (!FHndl) {
	 	FHndl = CreateWaitableTimer(
    	0,	// optional security attributes
			false,	// flag for manual-reset timer
	    name.c_str() 	// name of timer object
		);
   	Win32Check(FHndl != 0);
  }
}
//---------------------------------------------------------------------------
TWaitableTimer::TWaitableTimer() {
 	FHndl = CreateWaitableTimer(
   	0,	// optional security attributes
		false,	// flag for manual-reset timer
    0 	// name of timer object
	);
 	Win32Check(FHndl != 0);
}
//---------------------------------------------------------------------------
TWaitableTimer::TWaitableTimer(AnsiString name, TDateTime time, TDateTime interval) {
 	FHndl = OpenWaitableTimer(
    TIMER_ALL_ACCESS,	// access mode
    true,	// inherit flag
    name.c_str() 	// pointer to name of timer object
	);
	if (!FHndl) {
	 	FHndl = CreateWaitableTimer(
    	0,	// optional security attributes
			false,	// flag for manual-reset timer
	    name.c_str() 	// name of timer object
		);
   	Win32Check(FHndl != 0);
  }
  Set(time, interval);
}
//---------------------------------------------------------------------------
TWaitableTimer::TWaitableTimer(TDateTime time, TDateTime interval) {
 	FHndl = CreateWaitableTimer(
   	0,	// optional security attributes
		false,	// flag for manual-reset timer
    0 	// name of timer object
	);
 	Win32Check(FHndl != 0);
  Set(time, interval);
}
//---------------------------------------------------------------------------
void
TWaitableTimer::Set(TFileTime *time, int interval) {
	Win32Check(SetWaitableTimer(
    Hndl, //Hndl hTimer,	// Hndl to a timer object
    reinterpret_cast<LARGE_INTEGER*>(time), //const LARGE_INTEGER *pDueTime,	// when timer will become signaled
    interval, //LONG lPeriod,	// periodic timer interval
    0, //PTIMERAPCROUTINE pfnCompletionRoutine,	// pointer to the completion routine
    0, // LPVOID lpArgToCompletionRoutine,	// data passed to the completion routine
    0 //BOOL fResume	// flag for resume state
  ));
}
//---------------------------------------------------------------------------
void
TWaitableTimer::Set(TDateTime time, TDateTime interval) {
  TFileTime ft;
  DateTimeToFileTime(time, ft);
	Win32Check(SetWaitableTimer(
    Hndl, //Hndl hTimer,	// Hndl to a timer object
    reinterpret_cast<LARGE_INTEGER*>(&ft), //const LARGE_INTEGER *pDueTime,	// when timer will become signaled
    MilliSecondSpan(TDateTime(0), interval), //LONG lPeriod,	// periodic timer interval
    0, //PTIMERAPCROUTINE pfnCompletionRoutine,	// pointer to the completion routine
    0, // LPVOID lpArgToCompletionRoutine,	// data passed to the completion routine
    0 //BOOL fResume	// flag for resume state
  ));
}
//---------------------------------------------------------------------------
void
TWaitableTimer::Cancel() {
	Win32Check(CancelWaitableTimer(Hndl));
}
//---------------------------------------------------------------------------
__fastcall TSyncThread::TSyncThread(TSyncObj *syncObj, void (__closure *function)(),
                                    bool createSuspended, int timeout)
  :	SyncObj(syncObj),
    StartFunction(0),
    Function(function),
    Idle(0),
    StopFunction(0),
    Timeout(timeout),
	  TThread(true) {
  if (!Function)
    throw Exception("No function specifed in TSyncThread::TSyncThread()");
	if (!createSuspended)
  	Resume();
}
//---------------------------------------------------------------------------
__fastcall TSyncThread::TSyncThread(TSyncObj *syncObj, void (__closure *function)(),
                                    void (__closure *idle)(),
                                    bool createSuspended, int timeout)
  :	SyncObj(syncObj),
    StartFunction(0),
    Function(function),
    Idle(idle),
    StopFunction(0),
    Timeout(timeout),
	  TThread(true) {
  if (!Function)
    throw Exception("No function specifed in TSyncThread::TSyncThread()");
	if (!createSuspended)
  	Resume();
}
//---------------------------------------------------------------------------
__fastcall TSyncThread::TSyncThread(TSyncObj *syncObj, void (__closure *function)(),
                                    void (__closure *start)(), void (__closure *stop)(),
                                    bool createSuspended, int timeout)
  :	SyncObj(syncObj),
    StartFunction(start),
    Function(function),
    Idle(0),
    StopFunction(stop),
    Timeout(timeout),
	  TThread(true) {
  if (!Function)
    throw Exception("No function specifed in TSyncThread::TSyncThread()");
	if (!createSuspended)
  	Resume();
}
//---------------------------------------------------------------------------
__fastcall TSyncThread::TSyncThread(TSyncObj *syncObj, void (__closure *function)(),
                                    void (__closure *idle)(),
                                    void (__closure *start)(), void (__closure *stop)(),
                                    bool createSuspended, int timeout)
  :	SyncObj(syncObj),
    StartFunction(start),
    Function(function),
    Idle(idle),
    StopFunction(stop),
    Timeout(timeout),
	  TThread(true) {
  if (!Function)
    throw Exception("No function specifed in TSyncThread::TSyncThread()");
	if (!createSuspended)
  	Resume();
}
//---------------------------------------------------------------------------
void __fastcall
TSyncThread::Execute() {
  if (StartFunction)
    StartFunction();
	while (!Terminated)
  	if (SyncObj->Wait(Timeout))
    	Function();
    else if (Idle)
    	Idle();
  if (StopFunction)
    StopFunction();
}
//---------------------------------------------------------------------------
__fastcall TCycleThread::TCycleThread(TDateTime cycle, void (__closure *function)(),
                                      bool createSuspended, int timeout)
	:	TSyncThread(new TWaitableTimer(::NextCycle(cycle, Time = Now()), 0),
  		function, true, timeout),
    Cycle(cycle)
{
  if (!Function)
    throw Exception("No function specifed in TCycleThread::TCycleThread()");
	if (!createSuspended)
  	Resume();
}
//---------------------------------------------------------------------------
__fastcall TCycleThread::TCycleThread(TDateTime cycle, void (__closure *function)(),
                                      void (__closure *start)(), void (__closure *stop)(),
                                      bool createSuspended, int timeout)
	:	TSyncThread(new TWaitableTimer(::NextCycle(cycle, Time = Now()), 0),
  		function, start, stop, true, timeout),
    Cycle(cycle)
{
  if (!Function)
    throw Exception("No function specifed in TCycleThread::TCycleThread()");
	if (!createSuspended)
  	Resume();
}
//---------------------------------------------------------------------------
__fastcall TCycleThread::TCycleThread(TDateTime cycle, void (__closure *function)(),
                                      void (__closure *idle)(),
                                      void (__closure *start)(), void (__closure *stop)(),
                                      bool createSuspended, int timeout)
	:	TSyncThread(new TWaitableTimer(::NextCycle(cycle, Time = Now()), 0),
  		function, idle, start, stop, true, timeout),
    Cycle(cycle)
{
  if (!Function)
    throw Exception("No function specifed in TCycleThread::TCycleThread()");
	if (!createSuspended)
  	Resume();
}
//---------------------------------------------------------------------------
__fastcall TCycleThread::~TCycleThread() {
		delete SyncObj;
}
//---------------------------------------------------------------------------
void __fastcall
TCycleThread::Execute() {
  if (StartFunction)
    StartFunction();
	while (!Terminated)
  	if (SyncObj->Wait(Timeout)) {
      CycleTime = NearCycle(Cycle, Time = Now());
      dynamic_cast<TWaitableTimer*>(SyncObj)->
          Set(CycleTime + Cycle, 0);
    	Function();
    } else if (Idle)
    	Idle();
  if (StopFunction)
    StopFunction();
}
//---------------------------------------------------------------------------
//---------------------------------------------------------------------------
__fastcall TScheduler::TScheduler(void (__closure *function)(),
    bool createSuspended, TDateTime cycle, TDateTime offset)
	:	TSyncThread(new TWaitableTimer(NextCycle(cycle, Now()) + offset, 0),
        function, true, 1000),
    Cycle(cycle),
    Offset(offset)
{
  if (!Function)
    throw Exception("No function specifed in TScheduler::TScheduler()");
	if (!createSuspended)
  	Resume();
}
//---------------------------------------------------------------------------
__fastcall TScheduler::TScheduler(void (__closure *function)(),
    void (__closure *idle)(), bool createSuspended, TDateTime cycle,
    TDateTime offset)
	:	TSyncThread(new TWaitableTimer(NextCycle(cycle, Now()) + offset, 0),
        function, idle, true, 1000),
    Cycle(cycle),
    Offset(offset)
{
  if (!Function)
    throw Exception("No function specifed in TScheduler::TScheduler()");
	if (!createSuspended)
  	Resume();
}
//---------------------------------------------------------------------------
__fastcall TScheduler::TScheduler(void (__closure *function)(),
    void (__closure *start)(), void (__closure *stop)(),
    bool createSuspended, TDateTime cycle, TDateTime offset)
	:	TSyncThread(new TWaitableTimer(NextCycle(cycle, Now()) + offset, 0),
        function, start, stop, true, 1000),
    Cycle(cycle),
    Offset(offset)
{
  if (!Function)
    throw Exception("No function specifed in TScheduler::TScheduler()");
	if (!createSuspended)
  	Resume();
}
//---------------------------------------------------------------------------
__fastcall TScheduler::TScheduler(void (__closure *function)(),
    void (__closure *idle)(), void (__closure *start)(),
    void (__closure *stop)(), bool createSuspended, TDateTime cycle,
    TDateTime offset)
	:	TSyncThread(new TWaitableTimer(NextCycle(cycle, Now()) + offset, 0),
        function, idle, start, stop, true, 1000),
    Cycle(cycle),
    Offset(offset)
{
  if (!Function)
    throw Exception("No function specifed in TScheduler::TScheduler()");
	if (!createSuspended)
  	Resume();
}
//---------------------------------------------------------------------------
__fastcall TScheduler::~TScheduler()
{
	delete SyncObj;
}
//---------------------------------------------------------------------------
void __fastcall TScheduler::Execute()
{
  if (StartFunction)
    StartFunction();
	while (!Terminated)
  	if (SyncObj->Wait(Timeout)) {
      dynamic_cast<TWaitableTimer*>(SyncObj)->Set(
          (double(Offset) / double(Cycle) > 0.5 ?
          LastCycle(Cycle, Now() + SecToTime(1)) :
          NextCycle(Cycle, Now() + SecToTime(1))) + Offset, 0);
    	Function();
    } else if (Idle)
      Idle();
  if (StopFunction)
    StopFunction();
}
//---------------------------------------------------------------------------
//---------------------------------------------------------------------------
__fastcall TMHDScheduler::TMHDScheduler(int h, int m)
	:	TCycleThread(double(1) / 24 / 60, &OnCycle, true, 100),
  	H(h),
    M(m)
{
	Resume();
}
//---------------------------------------------------------------------------
__fastcall TMHDScheduler::~TMHDScheduler()
{
}
//---------------------------------------------------------------------------
void TMHDScheduler::OnCycle()
{
	OnMin();
  int m = MinuteOf(CycleTime);
  int h = HourOf(CycleTime);
  if (!m) {
  	OnHour();
  	if (!h)
    	OnDay();
  }
  if (m == M && h == H)
  	OnDayOffset();
}
//---------------------------------------------------------------------------
//---------------------------------------------------------------------------
__fastcall TEventList::TEventList(TStrings *sl) {
 	for (int i = 0; i < sl->Count; ++i)
  	Add(new TEvent(sl->Strings[i]));
}
//---------------------------------------------------------------------------
void __fastcall TEventList::Set() {
 	for (int i = 0; i < Count; ++i) {
  	TEvent *e = dynamic_cast<TEvent*>(Items[i]);
    if (!e)
    	throw Exception("Not TEvent object in TEventList::Set()");
    e->Set();
  }
}
//---------------------------------------------------------------------------
__fastcall TMultiSyncThread::TMultiSyncThread(TSyncList *sl, void (__closure *function)(int),
                                              bool createSuspended, bool all, int timeout)
	:	TThread(true),
		Count(sl->Count),
    StartFunction(0),
    Function(function),
    Idle(0),
    StopFunction(0),
    All(all),
    Timeout(timeout) {
  if (!Function)
    throw Exception("No function specifed in TMultiSyncThread::TMultiSyncThread()");
  SyncList = sl;
 	Hndl = new HANDLE[Count];
  for (int i = 0; i < Count; ++i)
  	Hndl[i] = SyncList->Items[i]->Hndl;
	if (!createSuspended)
  	Resume();
}
//---------------------------------------------------------------------------
__fastcall TMultiSyncThread::TMultiSyncThread(TSyncList *sl, void (__closure *function)(int),
                                              void (__closure *idle)(),
                                              bool createSuspended, bool all, int timeout)
	:	TThread(true),
		Count(sl->Count),
    StartFunction(0),
    Function(function),
    Idle(idle),
    StopFunction(0),
    All(all),
    Timeout(timeout) {
  if (!Function)
    throw Exception("No function specifed in TMultiSyncThread::TMultiSyncThread()");
  SyncList = sl;
 	Hndl = new HANDLE[Count];
  for (int i = 0; i < Count; ++i)
  	Hndl[i] = SyncList->Items[i]->Hndl;
	if (!createSuspended)
  	Resume();
}
//---------------------------------------------------------------------------
__fastcall TMultiSyncThread::TMultiSyncThread(TSyncList *sl, void (__closure *function)(int),
                                              void (__closure *start)(), void (__closure *stop)(),
                                              bool createSuspended, bool all, int timeout)
	:	TThread(true),
		Count(sl->Count),
    StartFunction(start),
    Function(function),
    Idle(0),
    StopFunction(stop),
    All(all),
    Timeout(timeout) {
  if (!Function)
    throw Exception("No function specifed in TMultiSyncThread::TMultiSyncThread()");
  SyncList = sl;
 	Hndl = new HANDLE[Count];
  for (int i = 0; i < Count; ++i)
  	Hndl[i] = SyncList->Items[i]->Hndl;
	if (!createSuspended)
  	Resume();
}
//---------------------------------------------------------------------------
__fastcall TMultiSyncThread::TMultiSyncThread(TSyncList *sl, void (__closure *function)(int),
                                              void (__closure *idle)(),
                                              void (__closure *start)(), void (__closure *stop)(),
                                              bool createSuspended, bool all, int timeout)
	:	TThread(true),
		Count(sl->Count),
    StartFunction(start),
    Function(function),
    Idle(idle),
    StopFunction(stop),
    All(all),
    Timeout(timeout) {
  if (!Function)
    throw Exception("No function specifed in TMultiSyncThread::TMultiSyncThread()");
  SyncList = sl;
 	Hndl = new HANDLE[Count];
  for (int i = 0; i < Count; ++i)
  	Hndl[i] = SyncList->Items[i]->Hndl;
	if (!createSuspended)
  	Resume();
}
//---------------------------------------------------------------------------
__fastcall TMultiSyncThread::~TMultiSyncThread() {
 	delete[] Hndl;
}
//---------------------------------------------------------------------------
void __fastcall
TMultiSyncThread::Execute() {
  if (StartFunction)
    StartFunction();
	while (!Terminated) {
	  DWORD ret = Count > 1 ?
	  	WaitForMultipleObjects(
  		  Count,	// number of Hndls in the object Hndl array
    	  Hndl,	// pointer to the object-Hndl array
	      All,	// wait all flag
  	    Timeout 	// time-out interval in milliseconds
      ) :
	  	WaitForSingleObject(
    	  Hndl[0],	// pointer to the object-Hndl array
  	    Timeout 	// time-out interval in milliseconds
      );
		if (ret == WAIT_TIMEOUT) {
      if (Idle)
    	  Idle();
    	continue;
    }
		if (ret == WAIT_FAILED)
   		RaiseLastOSError();
	  for (int i = 0;; ++i) {
	    if (i == Count)
  	 		throw(Exception("Unknown WaitForMultipleObjects() state"));
			if (ret == WAIT_OBJECT_0 + i) {
        TChangeNotification *cn = dynamic_cast<TChangeNotification*>(SyncList->Items[i]);
        if (cn)
        	cn->Next();
  	  	Function(i);
        break;
      }
	  	if (ret == WAIT_ABANDONED_0 + i)
				break;
	  }
  }
  if (StopFunction)
    StopFunction();
}
//---------------------------------------------------------------------------
} //namespace IPC
//---------------------------------------------------------------------------

