//---------------------------------------------------------------------------
/*
Copyright (c) 2013 Serge L. Ryadkow admin@opcgate.ru http://opcgate.ru 

Permission is hereby granted, free of charge, to any person obtaining a copy of this 
software and associated documentation files (the "Software"), to deal in the Software 
without restriction, including without limitation the rights to use, copy, modify, 
merge, publish, distribute, sublicense, and/or sell copies of the Software, and to 
permit persons to whom the Software is furnished to do so, subject to the following 
conditions:

The above copyright notice and this permission notice shall be included in all copies 
or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, 
INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR 
PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE 
FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR 
OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR 
OTHER DEALINGS IN THE SOFTWARE. 
*/
//---------------------------------------------------------------------------
//---------------------------------------------------------------------------
#ifndef IpcH
#define IpcH
#include "ListUtils.h"
//---------------------------------------------------------------------------
namespace IPC {
//---------------------------------------------------------------------------
class THndl {
 	public:
  	THndl() : FHndl(0) {};
  	virtual ~THndl();
		__property HANDLE Hndl  = { read=FHndl };
    operator HANDLE() {return Hndl;}
  protected:
  	HANDLE FHndl;
};
//---------------------------------------------------------------------------
class TSharedMemory : public THndl {
 	public:
  	TSharedMemory(AnsiString name, int size);
  	TSharedMemory(int size);
  	TSharedMemory(THndl *file, AnsiString name, int size);
  	TSharedMemory(THndl *file, int size);
  	virtual ~TSharedMemory();
		__property void * Ptr  = { read=FPtr };
		__property bool Existing  = { read=FExisting };
  protected:
  	void *FPtr;
    bool FExisting;
private:
};
//---------------------------------------------------------------------------
class TSyncObj : public THndl {
	public:
  	virtual bool Wait(int timeout = INFINITE);
};
//---------------------------------------------------------------------------
// not work with TChangeNotification class !!!!!!!!!!!!!!!!
TSyncObj *WaitAny(TSyncObj *o1, TSyncObj *o2, int timeout = INFINITE);
/*
TSyncObj *WaitAny(TList *l, int timeout = INFINITE);
TSyncObj *WaitAny(TList *l, TSyncObj *o, int timeout = INFINITE);
*/
//---------------------------------------------------------------------------
class TMutex : public TSyncObj {
 	public:
  	TMutex(AnsiString name);
  	TMutex();
    void Release();
  public:
  	class TLock {
     	public:
      	TLock(TMutex *mutex, int timeout = INFINITE);
        ~TLock();
      private:
      	TMutex *Mutex;
  	};
 	friend class TLock;
};
//---------------------------------------------------------------------------
class TEvent : public TSyncObj {
 	public:
  	TEvent(AnsiString name);
  	TEvent();
    void Set();
    void Reset();
};
//---------------------------------------------------------------------------
const FILE_NOTIFY_CHANGE_ANY = FILE_NOTIFY_CHANGE_FILE_NAME |
                               FILE_NOTIFY_CHANGE_DIR_NAME |
                               FILE_NOTIFY_CHANGE_ATTRIBUTES |
                               FILE_NOTIFY_CHANGE_SIZE |
                               FILE_NOTIFY_CHANGE_LAST_WRITE |
                               FILE_NOTIFY_CHANGE_SECURITY;
//---------------------------------------------------------------------------
class TChangeNotification : public TSyncObj {
 	public:
  	TChangeNotification(AnsiString path, bool subtree = true,
                        int filter = FILE_NOTIFY_CHANGE_ANY);
  	virtual ~TChangeNotification();
  	virtual bool Wait(int timeout = INFINITE);
  protected:
  	void Next();
	friend class TMultiSyncThread;
};
//---------------------------------------------------------------------------
class TWaitableTimer : public TSyncObj {
 	public:
  	TWaitableTimer(AnsiString name);
  	TWaitableTimer();
  	TWaitableTimer(AnsiString name, TDateTime time, TDateTime interval = TDateTime(.0));
  	TWaitableTimer(TDateTime time, TDateTime interval = TDateTime(.0));
    void Set(TFileTime *time, int interval = 0);
    void Set(TDateTime time, TDateTime interval = TDateTime(.0));
		void Cancel();
};
//---------------------------------------------------------------------------
class TSyncThread : public TThread {
	public:
		__fastcall TSyncThread(TSyncObj *syncObj, void (__closure *function)(),
                           bool createSuspended, int timeout = INFINITE);
		__fastcall TSyncThread(TSyncObj *syncObj, void (__closure *function)(),
                           void (__closure *idle)(),
                           bool createSuspended, int timeout = INFINITE);
		__fastcall TSyncThread(TSyncObj *syncObj, void (__closure *function)(),
                           void (__closure *start)(), void (__closure *stop)(),
                           bool createSuspended, int timeout = INFINITE);
		__fastcall TSyncThread(TSyncObj *syncObj, void (__closure *function)(),
                           void (__closure *idle)(),
                           void (__closure *start)(), void (__closure *stop)(),
                           bool createSuspended, int timeout = INFINITE);
	protected:
		virtual void __fastcall Execute();
  	void (__closure *Function)();
  	void (__closure *Idle)();
  	void (__closure *StartFunction)();
  	void (__closure *StopFunction)();
    TSyncObj *SyncObj;
    int Timeout;
};
//---------------------------------------------------------------------------
class TCycleThread : public TSyncThread {
	public:
		__fastcall TCycleThread(TDateTime cycle, void (__closure *function)(),
                            bool createSuspended, int timeout = INFINITE);
		__fastcall TCycleThread(TDateTime cycle, void (__closure *function)(),
                            void (__closure *start)(), void (__closure *stop)(),
                            bool createSuspended, int timeout = INFINITE);
		__fastcall TCycleThread(TDateTime cycle, void (__closure *function)(),
                            void (__closure *idle)(),
                            void (__closure *start)(), void (__closure *stop)(),
                            bool createSuspended, int timeout = INFINITE);
		virtual __fastcall ~TCycleThread();
    const TDateTime Cycle;
	protected:
		virtual void __fastcall Execute();
    TDateTime Time, CycleTime;
};
//---------------------------------------------------------------------------
class TScheduler : public TSyncThread {
public:
	__fastcall TScheduler(void (__closure *function)(),
      bool createSuspended, TDateTime cycle, TDateTime offset = 0);
	__fastcall TScheduler(void (__closure *function)(), void (__closure *idle)(),
      bool createSuspended, TDateTime cycle, TDateTime offset = 0);
	__fastcall TScheduler(void (__closure *function)(),
      void (__closure *start)(), void (__closure *stop)(),
      bool createSuspended, TDateTime cycle, TDateTime offset = 0);
	__fastcall TScheduler(void (__closure *function)(), void (__closure *idle)(),
      void (__closure *start)(), void (__closure *stop)(),
      bool createSuspended, TDateTime cycle, TDateTime offset = 0);
  virtual __fastcall ~TScheduler();
protected:
	virtual void __fastcall Execute();
  TDateTime Cycle, Offset;
};
//---------------------------------------------------------------------------
class TMHDScheduler : public TCycleThread {
public:
	__fastcall TMHDScheduler(int h = 0, int m = 0);
	__fastcall ~TMHDScheduler();
  const int H, M;
protected:
	virtual void __fastcall OnMin() {}
	virtual void __fastcall OnHour() {}
	virtual void __fastcall OnDay() {}
	virtual void __fastcall OnDayOffset() {}
private:
	void OnCycle();
};
//---------------------------------------------------------------------------
typedef TPtrList<TSyncObj> TSyncList;
//---------------------------------------------------------------------------
class TEventList : public TSyncList {
	public:
  	__fastcall TEventList() {}
  	__fastcall TEventList(TStrings *sl);
		void __fastcall Set();
};
//---------------------------------------------------------------------------
class TMultiSyncThread : public TThread {
	public:
		__fastcall TMultiSyncThread(TSyncList *sl, void (__closure *function)(int),
                                bool createSuspended, bool all, int timeout = INFINITE);
		__fastcall TMultiSyncThread(TSyncList *sl, void (__closure *function)(int),
                                void (__closure *idle)(),
                                bool createSuspended, bool all, int timeout = INFINITE);
		__fastcall TMultiSyncThread(TSyncList *sl, void (__closure *function)(int),
                                void (__closure *start)(), void (__closure *stop)(),
                                bool createSuspended, bool all, int timeout = INFINITE);
		__fastcall TMultiSyncThread(TSyncList *sl, void (__closure *function)(int),
                                void (__closure *idle)(),
                                void (__closure *start)(), void (__closure *stop)(),
                                bool createSuspended, bool all, int timeout = INFINITE);
		__fastcall ~TMultiSyncThread();
	protected:
		virtual void __fastcall Execute();
  	void (__closure *Function)(int);
  	void (__closure *Idle)();
  	void (__closure *StartFunction)();
  	void (__closure *StopFunction)();
    HANDLE *Hndl;
    int Timeout;
    int Count;
    bool All;
    TSyncList *SyncList;
};
//---------------------------------------------------------------------------
template <class T>
class TSharedArray {
  public:
  	__fastcall TSharedArray(AnsiString name, int count);
    __fastcall ~TSharedArray();
		__property T *Array  = { read=FArray };
		__property int Count  = { read=GetCount };
	private:
		T* FArray;
    int __fastcall GetCount();
    TMutex *Mutex;
    TSharedMemory *SharedMemory;
  public:
  	class TLock {
     	public:
      	__fastcall TLock(TSharedArray<T> *sa, int timeout = INFINITE);
        __fastcall ~TLock();
      private:
      	TMutex *Mutex;
  	};
 	friend class TLock;
};
//---------------------------------------------------------------------------
template <class T>
class TSharedObject {
  public:
  	__fastcall TSharedObject(AnsiString name, void *param);
    __fastcall ~TSharedObject();
		__property T *Object  = { read=FObject };
	private:
    TMutex *Mutex;
    TSharedMemory *SharedMemory;
    T *FObject;
  public:
  	class TLock {
     	public:
      	__fastcall TLock(TSharedObject<T> *so, int timeout = INFINITE);
        __fastcall ~TLock();
      private:
      	TMutex *Mutex;
  	};
 	friend class TLock;
};
//---------------------------------------------------------------------------
//---------------------------------------------------------------------------
// Template implementation
//---------------------------------------------------------------------------
//---------------------------------------------------------------------------
template<class T>
__fastcall TSharedArray<T>::TSharedArray(AnsiString name, int count) {
	Mutex = new TMutex(name + "Mutex");
	SharedMemory = new TSharedMemory(name + "SharedMemory",
                                   count * sizeof(T) + sizeof(int));
	if (SharedMemory->Existing) {
		FArray = (reinterpret_cast<T*>(
    	reinterpret_cast<char*>(SharedMemory->Ptr) +
      sizeof(int)));
  } else {
		FArray = new(reinterpret_cast<void*>(
    	reinterpret_cast<char*>(SharedMemory->Ptr) +
      sizeof(int))) T[count];
		*reinterpret_cast<int*>(SharedMemory->Ptr) = count * sizeof(T);
  }
}
//---------------------------------------------------------------------------
template<class T>
__fastcall TSharedArray<T>::~TSharedArray() {
	delete SharedMemory;
  delete Mutex;
}
//---------------------------------------------------------------------------
template<class T> int
__fastcall TSharedArray<T>::GetCount()
{
	return *reinterpret_cast<int*>(SharedMemory->Ptr) / sizeof(T);
}
//---------------------------------------------------------------------------
template<class T>
__fastcall TSharedArray<T>::TLock::TLock(TSharedArray<T> *sa, int timeout)
 	:	Mutex(sa->Mutex) {
	if (!Mutex->Wait(timeout))
   	throw Exception("Error locking shared array");
}
//---------------------------------------------------------------------------
template<class T>
__fastcall TSharedArray<T>::TLock::~TLock() {
	Mutex->Release();
}
//---------------------------------------------------------------------------
template<class T>
__fastcall TSharedObject<T>::TSharedObject(AnsiString name, void *param) {
	Mutex = new TMutex(name + "Mutex");
	SharedMemory = new TSharedMemory(name + "SharedMemory", sizeof(T));
	if (SharedMemory->Existing)
		FObject = reinterpret_cast<T*>(SharedMemory->Ptr);
  else
		FObject = new(SharedMemory->Ptr) T(param);
}
//---------------------------------------------------------------------------
template<class T>
__fastcall TSharedObject<T>::~TSharedObject() {
	delete SharedMemory;
  delete Mutex;
}
//---------------------------------------------------------------------------
template<class T>
__fastcall TSharedObject<T>::TLock::TLock(TSharedObject<T> *sa, int timeout)
 	:	Mutex(sa->Mutex) {
	if (!Mutex->Wait(timeout))
   	throw Exception("Error locking shared array");
}
//---------------------------------------------------------------------------
template<class T>
__fastcall TSharedObject<T>::TLock::~TLock() {
	Mutex->Release();
}
//---------------------------------------------------------------------------
} //namespace IPC
using namespace IPC;
//---------------------------------------------------------------------------
#endif
