//---------------------------------------------------------------------------
/*
Copyright (c) 2013 Serge L. Ryadkow admin@opcgate.ru http://opcgate.ru 

Permission is hereby granted, free of charge, to any person obtaining a copy of this 
software and associated documentation files (the "Software"), to deal in the Software 
without restriction, including without limitation the rights to use, copy, modify, 
merge, publish, distribute, sublicense, and/or sell copies of the Software, and to 
permit persons to whom the Software is furnished to do so, subject to the following 
conditions:

The above copyright notice and this permission notice shall be included in all copies 
or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, 
INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR 
PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE 
FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR 
OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR 
OTHER DEALINGS IN THE SOFTWARE. 
*/
//---------------------------------------------------------------------------
//---------------------------------------------------------------------------
#include <vcl.h>

#pragma hdrstop

#include <DateUtils.hpp>
#include <StrUtils.hpp>

#include "JSONUtil.h"
#include "Util.h"
//---------------------------------------------------------------------------
#pragma package(smart_init)
//---------------------------------------------------------------------------
__fastcall TJsonMinMax::TJsonMinMax()
{
  MinMax = 0;
  Cnt = 0;
}
//---------------------------------------------------------------------------
__fastcall TJsonMinMax::~TJsonMinMax()
{
  delete[] MinMax;
}
//---------------------------------------------------------------------------
AnsiString __fastcall TJsonMinMax::ToJson()
{
  AnsiString s;
  for (int i = 0; i < Cnt; ++i) {
    if (i)
      s += ",";
    s += "\"" + MinMax[i].FN + "min\":" +
        (MinMax[i].F ? AnsiString(MinMax[i].Min) : AnsiString("null")) +
        ",\"" + MinMax[i].FN + "max\":" +
        (MinMax[i].F ? AnsiString(MinMax[i].Max) : AnsiString("null"));
  }
  return s;
}
//---------------------------------------------------------------------------
//---------------------------------------------------------------------------
void __fastcall TJsonMinMax::TMinMax::Set(double v)
{
  if (!F) {
    Min = Max = v;
    F = true;
  } else {
    if (v < Min)
      Min = v;
    if (v > Max)
      Max = v;
  }
}
//---------------------------------------------------------------------------
//---------------------------------------------------------------------------
AnsiString DataSet2Json(TDataSet *ds, TJsonTimeFmt tf, TJsonMinMax *mm,
    bool noProcVarStr)
{
  AnsiString s;
  bool ocf = !ds->Active;
  if (ocf)
    ds->Open();
  try {
    if (mm && !mm->Cnt) {
      for (int i = 0; i < ds->FieldCount; ++i)
        if (ds->Fields->Fields[i]->Visible)
          switch (ds->Fields->Fields[i]->DataType) {
            case ftSmallint:
            case ftInteger:
            case ftFloat:
            case ftBCD:
            case ftWord:
              ++mm->Cnt;
              break;
          }
      if (mm->Cnt) {
        mm->MinMax = new TJsonMinMax::TMinMax[mm->Cnt];
        for (int i = 0, j = 0; i < ds->FieldCount; ++i)
          if (ds->Fields->Fields[i]->Visible)
            switch (ds->Fields->Fields[i]->DataType) {
            case ftInteger:
            case ftSmallint:
            case ftLargeint:
            case ftAutoInc:
            case ftFloat:
            case ftBCD:
            case ftWord:
                mm->MinMax[j++].FN = ds->Fields->Fields[i]->DisplayLabel;
                break;
            }
      }
    }
    for(bool f = true; !ds->Eof; ds->Next()) {
      if (!f)
        s += ",";
      f = false;
      s += "{";
      for (int i = 0, j = 0, f = true; i < ds->FieldCount; ++i) {
        if (ds->Fields->Fields[i]->Visible) {
          if (!f)
            s += ",";
          f = false;
          s += "\"" + ds->Fields->Fields[i]->DisplayLabel + "\":";
          switch (ds->Fields->Fields[i]->DataType) {
            case ftInteger:
            case ftSmallint:
            case ftLargeint:
            case ftAutoInc:
            case ftFloat:
            case ftBCD:
            case ftWord:
              s += (ds->Fields->Fields[i]->IsNull ? AnsiString("null") :
                ds->Fields->Fields[i]->AsString);
              if (mm && j < mm->Cnt) {
                if (!ds->Fields->Fields[i]->IsNull)
                  mm->MinMax[j].Set(ds->Fields->Fields[i]->AsFloat);
                ++j;
              }
              break;
            case ftDateTime:
              switch(tf) {
                case tfLS:
                  s += (ds->Fields->Fields[i]->IsNull ? AnsiString("null") :
                    "\"" + FormatDateTime("dd\".\"mm\".\"yyyy\" \"hh\":\"nn\":\"ss",
                      ds->Fields->Fields[i]->AsDateTime) + "\"");
                  break;
                case tfGS:
                  s += (ds->Fields->Fields[i]->IsNull ? AnsiString("null") :
                    "\"" + FormatDateTime("dd\".\"mm\".\"yyyy\" \"hh\":\"nn\":\"ss",
                      ::LocalToGmt(ds->Fields->Fields[i]->AsDateTime)) + "\"");
                  break;
                case tfLI:
                  s += (ds->Fields->Fields[i]->IsNull ? AnsiString("null") :
                    AnsiString(::DateTimeToUnix(
                      ds->Fields->Fields[i]->AsDateTime)));
                  break;
                case tfGI:
                  s += (ds->Fields->Fields[i]->IsNull ? AnsiString("null") :
                    AnsiString(::DateTimeToUnix(::LocalToGmt(
                      ds->Fields->Fields[i]->AsDateTime))));
                  break;
              }
              break;
            case ftVariant:
              if (noProcVarStr) {
                s += (ds->Fields->Fields[i]->IsNull ? AnsiString("null") :
                  ds->Fields->Fields[i]->AsString);
                break;
              }
            default:
              s += "\"" + AnsiToUtf8(AnsiReplaceStr(
                ds->Fields->Fields[i]->AsString, "\"", "\\\"")) + "\"";
          }
        }
      }
      s += "}";
    }
  } __finally {
    if (ocf)
      ds->Close();
  }
  return s;
}
//---------------------------------------------------------------------------
