//---------------------------------------------------------------------------
/*
Copyright (c) 2013 Serge L. Ryadkow admin@opcgate.ru http://opcgate.ru 

Permission is hereby granted, free of charge, to any person obtaining a copy of this 
software and associated documentation files (the "Software"), to deal in the Software 
without restriction, including without limitation the rights to use, copy, modify, 
merge, publish, distribute, sublicense, and/or sell copies of the Software, and to 
permit persons to whom the Software is furnished to do so, subject to the following 
conditions:

The above copyright notice and this permission notice shall be included in all copies 
or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, 
INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR 
PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE 
FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR 
OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR 
OTHER DEALINGS IN THE SOFTWARE. 
*/
//---------------------------------------------------------------------------
//---------------------------------------------------------------------------
#ifndef ListUtilsH
#define ListUtilsH
//---------------------------------------------------------------------------
#include <Classes.hpp>
#include <Contnrs.hpp>
#include <SyncObjs.hpp>
//---------------------------------------------------------------------------
////////////////////////////////////////////////
//TPtrList is object owner: destructor deletes all objects !
////////////////////////////////////////////////
template <class T>
class TPtrList : public TList {
	public:
    virtual void __fastcall Clear();
    virtual void __fastcall Delete(int i) {delete Items[i]; TList::Delete(i);}
    T *__fastcall Extract(T *item) {return reinterpret_cast<T*>(TList::Extract(item));}
    T *__fastcall First() {return reinterpret_cast<T*>(TList::First()));}
    T *__fastcall Last() {return reinterpret_cast<T*>(TList::Last()));}
		int __fastcall Remove(T *item) {int i = IndexOf(item); Delete(i); return i;}
		void __fastcall Sort(bool desc = false);
    T *__fastcall Find(T *item);
		__property T *Items[int i]  = { read=GetItems, write=SetItems };
    void __fastcall LoadFromStream(TStream *s);
    void __fastcall SaveToStream(TStream *s);
    void __fastcall LoadFromFile(const AnsiString &fn);
    void __fastcall SaveToFile(const AnsiString &fn);
	private:
		T* __fastcall GetItems(int i) {return reinterpret_cast<T*>(TList::Items[i]);}
		void __fastcall SetItems(int i, T *value) {TList::Items[i] = reinterpret_cast<void*>(value);}
    static int __fastcall Compare1(void *item1, void *item2);
    static int __fastcall Compare2(void *item1, void *item2);
};
//---------------------------------------------------------------------------
template <class T>
class TTSList {
  public:
    __fastcall TTSList();
    __fastcall ~TTSList();
    TPtrList<T>* __fastcall Lock();
    void __fastcall Unlock();
  private:
    TPtrList<T> *List;
    TCriticalSection *CS;
};
//---------------------------------------------------------------------------
template <class T>
class TMREWList {
  public:
    __fastcall TMREWList();
    __fastcall ~TMREWList();
    TPtrList<T>* __fastcall LockR();
    TPtrList<T>* __fastcall LockW();
    void __fastcall UnlockR();
    void __fastcall UnlockW();
    __property unsigned RevisionLevel = {read = GetRevisionLevel};
  private:
    unsigned __fastcall GetRevisionLevel() {return Lock->RevisionLevel;}
    TPtrList<T> *List;
    TMREWSync *Lock;
};
//---------------------------------------------------------------------------
template <class T>
class TPtrQueue : public TQueue {
  public:
    T *__fastcall Peek() {return reinterpret_cast<T*>(TQueue::Peek());}
    T *__fastcall Pop() {return reinterpret_cast<T*>(TQueue::Pop());}
    T *__fastcall Push(T *item) {return reinterpret_cast<T*>(TQueue::Push(item));}
};
//---------------------------------------------------------------------------
template <class T>
class TSmallTypeList : public TList {
	public:
  	__fastcall TSmallTypeList();
    int __fastcall Add(T item) {return TList::Add(reinterpret_cast<void*>(item));}
    T __fastcall Extract(T item) {return reinterpret_cast<T>(TList::Extract(reinterpret_cast<void*>(item))));}
    T __fastcall First() {return reinterpret_cast<T>(TList::First()));}
    T __fastcall Last() {return reinterpret_cast<T>(TList::Last());}
    int __fastcall IndexOf(T item) {return TList::IndexOf(reinterpret_cast<void*>(item));}
    void __fastcall Insert(int index, T item) {TList::Insert(index, reinterpret_cast<void*>(item));}
    int __fastcall Remove(T item) {return TList::Remove(reinterpret_cast<void*>(item));}
		void __fastcall Sort(bool desc = false);
		__property T Items[int i]  = { read=GetItems, write=SetItems };
		__property T Max  = { read=GetMax };
		__property T Min  = { read=GetMin };
	private:
		void __fastcall SetItems(int i, T value) {TList::Items[i] = reinterpret_cast<void*>(value);}
		T __fastcall GetItems(int i) {return reinterpret_cast<T>(TList::Items[i]);}
    static int __fastcall Compare1(void *item1, void *item2);
    static int __fastcall Compare2(void *item1, void *item2);
    T __fastcall GetMax();
    T __fastcall GetMin();
};
//---------------------------------------------------------------------------
template <class T>
class TStringSmallTypeList : public TStringList {
	public:
  	__fastcall TStringSmallTypeList();
		virtual int __fastcall Add(const AnsiString s, T v) {return AddObject(s, reinterpret_cast<System::TObject*>(v));}
		virtual int __fastcall IndexOf(const AnsiString s) {return TStringList::IndexOf(s);}
		virtual int __fastcall IndexOf(T v) {return IndexOfObject(reinterpret_cast<System::TObject*>(v));}
    virtual void __fastcall Insert(int ind, const AnsiString s, T v) {InsertObject(ind, s, reinterpret_cast<System::TObject*>(v));}
		__property T Items[int i]  = { read=GetItems, write=SetItems };
	private:
		void __fastcall SetItems(int i, T value) {Objects[i] = reinterpret_cast<System::TObject*>(value);}
		T __fastcall GetItems(int i) {return reinterpret_cast<T>(Objects[i]);}
};
//---------------------------------------------------------------------------
template <class T>
class TFileParseList : public TPtrList<T>
{
  public:
    __fastcall TFileParseList(AnsiString fn, AnsiString bs = ";");
};
//---------------------------------------------------------------------------
//---------------------------------------------------------------------------
//---------------------------------------------------------------------------
// TPtrList implementation
//---------------------------------------------------------------------------
//---------------------------------------------------------------------------
template <class T>
void __fastcall TPtrList<T>::Clear() {
	for (int i = 0; i < Count; ++i)
  	delete Items[i];
  TList::Clear();
}
//---------------------------------------------------------------------------
template <class T>
void __fastcall TPtrList<T>::Sort(bool desc) {
  if (desc)
    TList::Sort(Compare2);
  else
    TList::Sort(Compare1);
}
//---------------------------------------------------------------------------
template <class T>
T *__fastcall TPtrList<T>::Find(T *item) {
  for (int i = 0; i < Count; ++i)
  	if (*Items[i] == *item)
      return Items[i];
  return 0;
}
//---------------------------------------------------------------------------
template <class T>
int __fastcall TPtrList<T>::Compare1(void *item1, void *item2) {
	if (*reinterpret_cast<T*>(item1) > *reinterpret_cast<T*>(item2))
  	return 1;
	else if (*reinterpret_cast<T*>(item1) < *reinterpret_cast<T*>(item2))
  	return -1;
  else
  	return 0;
}
//---------------------------------------------------------------------------
template <class T>
int __fastcall TPtrList<T>::Compare2(void *item1, void *item2) {
	if (*reinterpret_cast<T*>(item1) < *reinterpret_cast<T*>(item2))
  	return 1;
	else if (*reinterpret_cast<T*>(item1) > *reinterpret_cast<T*>(item2))
  	return -1;
  else
  	return 0;
}
//---------------------------------------------------------------------------
template <class T>
void __fastcall TPtrList<T>::LoadFromStream(TStream *s) {
  for (;;) {
    T *r = new T;
    int l = s->Read(r, sizeof(T));
    if (l == sizeof(T))
      Add(r);
    else {
      delete r;
      if (l)
        throw Exception("TPtrList<T>::LoadFromStream() Error");
      break;
    }
  }
}
//---------------------------------------------------------------------------
template <class T>
void __fastcall TPtrList<T>::SaveToStream(TStream *s) {
  for (int i = 0; i < Count; ++i)
    if (s->Write(Items[i], sizeof(T)) != sizeof(T))
      throw Exception("TPtrList<T>::SaveToStream() Error");
}
//---------------------------------------------------------------------------
template <class T>
void __fastcall TPtrList<T>::LoadFromFile(const AnsiString &fn) {
  TFileStream *f = new TFileStream(fn, fmOpenRead	| fmShareDenyWrite);
  try {
    LoadFromStream(f);
  } __finally {
    delete f;
  }
}
//---------------------------------------------------------------------------
template <class T>
void __fastcall TPtrList<T>::SaveToFile(const AnsiString &fn) {
  TFileStream *f = new TFileStream(fn,
    (FileExists(fn) ? fmOpenReadWrite : fmCreate)	| fmShareExclusive);
  try {
    SaveToStream(f);
  } __finally {
    delete f;
  }
}
//---------------------------------------------------------------------------
//---------------------------------------------------------------------------
// TTSList implementation
//---------------------------------------------------------------------------
//---------------------------------------------------------------------------
template <class T>
__fastcall TTSList<T>::TTSList() {
  List = new TPtrList<T>();
  CS = new TCriticalSection();
}
//---------------------------------------------------------------------------
template <class T>
__fastcall TTSList<T>::~TTSList() {
  delete List;
  delete CS;
}
//---------------------------------------------------------------------------
template <class T>
TPtrList<T>* __fastcall TTSList<T>::Lock() {
  CS->Enter();
  return List;
}
//---------------------------------------------------------------------------
template <class T>
void __fastcall TTSList<T>::Unlock() {
  CS->Leave();
}
//---------------------------------------------------------------------------
//---------------------------------------------------------------------------
// TMREWList implementation
//---------------------------------------------------------------------------
//---------------------------------------------------------------------------
template <class T>
__fastcall TMREWList<T>::TMREWList() {
  List = new TPtrList<T>();
  Lock = new TMREWSync();
}
//---------------------------------------------------------------------------
template <class T>
__fastcall TMREWList<T>::~TMREWList() {
  delete List;
  delete Lock;
}
//---------------------------------------------------------------------------
template <class T>
TPtrList<T>* __fastcall TMREWList<T>::LockR() {
  Lock->BeginRead();
  return List;
}
//---------------------------------------------------------------------------
template <class T>
TPtrList<T>* __fastcall TMREWList<T>::LockW() {
  Lock->BeginWrite();
  return List;
}
//---------------------------------------------------------------------------
template <class T>
void __fastcall TMREWList<T>::UnlockR() {
  Lock->EndRead();
}
//---------------------------------------------------------------------------
template <class T>
void __fastcall TMREWList<T>::UnlockW() {
  Lock->EndWrite();
}
//---------------------------------------------------------------------------
//---------------------------------------------------------------------------
// TSmallTypeList implementation
//---------------------------------------------------------------------------
//---------------------------------------------------------------------------
#pragma warn -8008
#pragma warn -8066
//---------------------------------------------------------------------------
template <class T>
__fastcall TSmallTypeList<T>::TSmallTypeList()
	: TList() {
  if (sizeof(T) > sizeof(void*))
  	throw Exception("sizeof(T) > sizeof(void*) in template <class T> __fastcall TSmallTypeList<T>::TSmallTypeList()");
}
//---------------------------------------------------------------------------
#pragma warn +8066
#pragma warn +8008
//---------------------------------------------------------------------------
template <class T>
void __fastcall TSmallTypeList<T>::Sort(bool desc)
{
  if (desc)
    TList::Sort(Compare2);
  else
    TList::Sort(Compare1);
}
//---------------------------------------------------------------------------
template <class T>
int __fastcall TSmallTypeList<T>::Compare1(void *item1, void *item2) {
	if (reinterpret_cast<T>(item1) > reinterpret_cast<T>(item2))
  	return 1;
	if (reinterpret_cast<T>(item1) < reinterpret_cast<T>(item2))
  	return -1;
  else
  	return 0;
}
//---------------------------------------------------------------------------
template <class T>
int __fastcall TSmallTypeList<T>::Compare2(void *item1, void *item2) {
	if (reinterpret_cast<T>(item1) < reinterpret_cast<T>(item2))
  	return 1;
	if (reinterpret_cast<T>(item1) > reinterpret_cast<T>(item2))
  	return -1;
  else
  	return 0;
}
//---------------------------------------------------------------------------
template <class T>
T __fastcall TSmallTypeList<T>::GetMax() {
  T max = Count ? Items[0] : 0;
  for (int i = 1; i < Count; ++i)
		if (max < Items[i])
			max = Items[i];
 	return max;
}
//---------------------------------------------------------------------------
template <class T>
T __fastcall TSmallTypeList<T>::GetMin() {
  T min = Count ? Items[0] : 0;
  for (int i = 1; i < Count; ++i)
		if (min > Items[i])
			min = Items[i];
 	return min;
}
//---------------------------------------------------------------------------
//---------------------------------------------------------------------------
// TStringSmallTypeList implementation
//---------------------------------------------------------------------------
//---------------------------------------------------------------------------
#pragma warn -8008
#pragma warn -8066
//---------------------------------------------------------------------------
template <class T>
__fastcall TStringSmallTypeList<T>::TStringSmallTypeList()
	: TStringList() {
  if (sizeof(T) > sizeof(void*))
  	throw Exception("sizeof(T) > sizeof(void*) in template <class T> __fastcall TStringSmallTypeList<T>::TStringSmallTypeList()");
}
//---------------------------------------------------------------------------
#pragma warn +8066
#pragma warn +8008
//---------------------------------------------------------------------------
//---------------------------------------------------------------------------
// TFileParseTypeList implementation
//---------------------------------------------------------------------------
//---------------------------------------------------------------------------
template <class T>
__fastcall TFileParseList<T>::TFileParseList(AnsiString fn, AnsiString bs)
{
  TStringList *sl = new TStringList;
  try {
    sl->LoadFromFile(fn);
    RemoveComments(sl, bs);
    for (int i = 0; i < sl->Count; ++i)
      Add(new T(sl->Strings[i]));
  }
  __finally {
    delete sl;
  }
}
//---------------------------------------------------------------------------
//---------------------------------------------------------------------------
typedef TSmallTypeList<char> TCharList;
typedef TSmallTypeList<unsigned char> TUCharList;
typedef TSmallTypeList<signed char> TSCharList;
typedef TSmallTypeList<int> TIntList;
typedef TSmallTypeList<unsigned int> TUIntList;
typedef TSmallTypeList<signed int> TSIntList;
typedef TSmallTypeList<long> TLongList;
typedef TSmallTypeList<unsigned long> TULongList;
typedef TSmallTypeList<signed long> TSLongList;
typedef TSmallTypeList<short> TShortList;
typedef TSmallTypeList<unsigned short> TUShortList;
typedef TSmallTypeList<signed short> TSShortList;
typedef TSmallTypeList<float> TFloatList;
typedef TStringSmallTypeList<char> TStringCharList;
typedef TStringSmallTypeList<unsigned char> TStringUCharList;
typedef TStringSmallTypeList<signed char> TStringSCharList;
typedef TStringSmallTypeList<int> TStringIntList;
typedef TStringSmallTypeList<unsigned int> TStringUIntList;
typedef TStringSmallTypeList<signed int> TStringSIntList;
typedef TStringSmallTypeList<long> TStringLongList;
typedef TStringSmallTypeList<unsigned long> TStringULongList;
typedef TStringSmallTypeList<signed long> TStringSLongList;
typedef TStringSmallTypeList<short> TStringShortList;
typedef TStringSmallTypeList<unsigned short> TStringUShortList;
typedef TStringSmallTypeList<signed short> TStringSShortList;
typedef TStringSmallTypeList<float> TStringFloatList;
//---------------------------------------------------------------------------
#endif
