//---------------------------------------------------------------------------
/*
Copyright (c) 2013 Serge L. Ryadkow admin@opcgate.ru http://opcgate.ru 

Permission is hereby granted, free of charge, to any person obtaining a copy of this 
software and associated documentation files (the "Software"), to deal in the Software 
without restriction, including without limitation the rights to use, copy, modify, 
merge, publish, distribute, sublicense, and/or sell copies of the Software, and to 
permit persons to whom the Software is furnished to do so, subject to the following 
conditions:

The above copyright notice and this permission notice shall be included in all copies 
or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, 
INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR 
PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE 
FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR 
OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR 
OTHER DEALINGS IN THE SOFTWARE. 
*/
//---------------------------------------------------------------------------
//---------------------------------------------------------------------------
#include <vcl.h>

#pragma hdrstop

#include <winbase.h>

#include "ProcMeter.h"
//---------------------------------------------------------------------------
#pragma package(smart_init)
//---------------------------------------------------------------------------
__fastcall TProcMeter::TProcMeter(int cycleMS, bool waitTerm)
	: TThread(true)
{
	CS = new TCriticalSection();
	WaitTerm = waitTerm;
  TermF = false;
  if (cycleMS > 100) {
  	Cycle = 100;
    Mul = cycleMS / 100;
  } else {
	  Cycle = cycleMS;
  	Mul = 0;
  }
	SYSTEM_INFO si;
	GetSystemInfo(&si);
  NumProc = si.dwNumberOfProcessors;
  PH = ::GetCurrentProcess();
  Measure();
  Measure();
  Resume();
}
//---------------------------------------------------------------------------
__fastcall TProcMeter::~TProcMeter()
{
	if (!TermF) {
		TThread::Terminate();
  	WaitFor();
  } else if (!WaitTerm)
  	WaitFor();
  delete CS;
}
//---------------------------------------------------------------------------
void __fastcall TProcMeter::Terminate()
{
	TThread::Terminate();
  if (WaitTerm)
  	WaitFor();
  TermF = true;
}
//---------------------------------------------------------------------------
double __fastcall TProcMeter::GetPercent()
{
	__int64 dt, t;
	CS->Enter();
  try {
  	dt = Last.ST - Prev.ST;
    t = (Last.KT - Prev.KT) + (Last.UT - Prev.UT);
  } __finally {
  	CS->Leave();
  }
  return dt ? double(t) / double(dt) * 100 / NumProc : 0;
}
//---------------------------------------------------------------------------
void __fastcall TProcMeter::Execute()
{
	while (!Terminated) {
		if (Mul)
    	for (int i = 0; i < Mul && !Terminated; ++i)
	  		::Sleep(Cycle);
    else
	  	::Sleep(Cycle);
    if (!Terminated)
	    Measure();
  }
}
//---------------------------------------------------------------------------
void __fastcall TProcMeter::Measure()
{
	CS->Enter();
  try {
  	memcpy(&Prev, &Last, sizeof(Last));
    GetProcessTimes(PH,
      (FILETIME*)(void*)&Last.CT,
      (FILETIME*)(void*)&Last.ET,
      (FILETIME*)(void*)&Last.KT,
      (FILETIME*)(void*)&Last.UT
    );
  	GetSystemTimeAsFileTime((FILETIME*)(void*)&Last.ST);
  } __finally {
  	CS->Leave();
  }
}
//---------------------------------------------------------------------------
