//---------------------------------------------------------------------------
/*
Copyright (c) 2013 Serge L. Ryadkow admin@opcgate.ru http://opcgate.ru 

Permission is hereby granted, free of charge, to any person obtaining a copy of this 
software and associated documentation files (the "Software"), to deal in the Software 
without restriction, including without limitation the rights to use, copy, modify, 
merge, publish, distribute, sublicense, and/or sell copies of the Software, and to 
permit persons to whom the Software is furnished to do so, subject to the following 
conditions:

The above copyright notice and this permission notice shall be included in all copies 
or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, 
INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR 
PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE 
FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR 
OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR 
OTHER DEALINGS IN THE SOFTWARE. 
*/
//---------------------------------------------------------------------------
//---------------------------------------------------------------------------

#include <vcl.h>

#pragma hdrstop

#include <Registry.hpp>
#include <StrUtils.hpp>
#include <Math.h>

#include "StringUtil.h"

//---------------------------------------------------------------------------

#pragma package(smart_init)
//---------------------------------------------------------------------------
AnsiString __fastcall Expand(const AnsiString &src, int length)
{
  if (src.Length() < length)
    return src + AnsiString::StringOfChar(' ', length - src.Length());
  else
    return src;
}
//---------------------------------------------------------------------------
AnsiString __fastcall Preconv(AnsiString s)
{
	return AnsiReplaceStr(s, "\t", " ").Trim();
}
//---------------------------------------------------------------------------
AnsiString __fastcall RemoveComments(AnsiString s, char b)
{
  AnsiString bs(b);
  int pos = s.Pos(bs);
  if (pos)
    s = s.SubString(1, pos-1);
  return s.Trim();
}
//---------------------------------------------------------------------------
AnsiString __fastcall RemoveComments(AnsiString s, AnsiString bs)
{
  int n = bs.Length();
  for (int i = 1; i <= n; ++i)
    s = RemoveComments(s, bs[i]);
  return s;
}
//---------------------------------------------------------------------------
void __fastcall RemoveComments(TStrings *sl, char b)
{
  for (int i = 0; i < sl->Count; ++i)
    sl->Strings[i] = RemoveComments(sl->Strings[i], b);
  RemoveEmptyStrings(sl);
}
//---------------------------------------------------------------------------
void __fastcall RemoveComments(TStrings *sl, AnsiString bs)
{
  for (int i = 0; i < sl->Count; ++i)
    sl->Strings[i] = RemoveComments(sl->Strings[i], bs);
  RemoveEmptyStrings(sl);
}
//---------------------------------------------------------------------------
void __fastcall RemoveEmptyStrings(TStrings *sl)
{
	while (sl->IndexOf("") > -1)
  	sl->Delete(sl->IndexOf(""));
}
//---------------------------------------------------------------------------
bool __fastcall FindEq(AnsiString s, AnsiString &res, AnsiString &expr)
{
	int i = s.AnsiPos("=");
  if (!i)
  	return false;
  if (i > 1) {
   	switch (s[i - 1]) {
     	case '!':
      case '<':
      case '>':
      	return false;
    }
  }
  if (i < s.Length() && s[i + 1] == '=')
  	return false;
  res = LeftStr(s, i - 1);
  expr = MidStr(s, i + 1, s.Length());
  return true;
}
//---------------------------------------------------------------------------
int __fastcall SplitString(const AnsiString &s, TStrings *sl, const AnsiString &ds)
{
  sl->Clear();
  AnsiString ss = s + "\n";
  for (int i = 1, j = 1; i <= ss.Length(); ++i) {
    if (ss.IsDelimiter(ds + "\n", i)) {
      sl->Append(ss.SubString(j, i - j));
      while (ss.IsDelimiter(ds, i))
        ++i;
      j = i;
    }
  }
  return sl->Count;
}
//---------------------------------------------------------------------------
AnsiString __fastcall GetPart(const AnsiString &s, int i, const AnsiString &ds)
{
  AnsiString ss;
	TStringList *sl = new TStringList;
  try {
  	ss = i < SplitString(s, sl, ds) ? sl->Strings[i] : AnsiString("");
  }
  __finally {
   	delete sl;
  }
  return ss;
}
//---------------------------------------------------------------------------
int __fastcall GetSubStrings(TStrings *src, TStrings *dst, int from, int to)
{
  dst->Clear();
  if (from >= 0 && from < src->Count && to >= 0 && to < src->Count)
    for (int i = from; i <= to; ++i)
      dst->Add(src->Strings[i]);
  return dst->Count;
}
//---------------------------------------------------------------------------
int __fastcall GetSubStrings(TStrings *src, TStrings *dst,
                             const AnsiString &from, const AnsiString &to)
{
  int i1 = src->IndexOf(from);
  int i2 = src->IndexOf(to);
  return GetSubStrings(src, dst, i1, i2);
}
//---------------------------------------------------------------------------
int __fastcall GetSection(TStrings *src, TStrings *dst,
                          const AnsiString &from, const AnsiString &to)
{
  int n = GetSubStrings(src, dst, from, to);
  if (n >= 2) {
    dst->Delete(n - 1);
    dst->Delete(0);
  } else
    dst->Clear();
  return dst->Count;
}
//---------------------------------------------------------------------------
int __fastcall GetSection(TStrings *sl,
                          const AnsiString &from, const AnsiString &to)
{
  int i1 = sl->IndexOf(from);
  int i2 = sl->IndexOf(to);
  if (i1 >= 0 && i2 >= 0 && i2 > i1 + 1) {
    while (sl->Count > i2)
      sl->Delete(i2);
    for (int i = 0; i <= i1; ++i)
      sl->Delete(0);
  } else
    sl->Clear();
  return sl->Count;
}
//---------------------------------------------------------------------------
int __fastcall
StringToIntList(AnsiString s, TIntList *il, AnsiString ds, AnsiString cds) {
	il->Clear();
	TStringList *sl1 = new TStringList;
	TStringList *sl2 = new TStringList;
  try {
		SplitString(s, sl1, cds);
    RemoveEmptyStrings(sl1);
    for (int i = 0; i < sl1->Count; ++i) {
			SplitString(sl1->Strings[i], sl2, ds);
	    RemoveEmptyStrings(sl2);
			if (i) {
       	int to = sl2->Strings[0].Trim().ToInt();
        for (int j = il->Last() + 1; j < to; ++j)
					il->Add(j);
      }
	    for (int j = 0; j < sl2->Count; ++j)
				il->Add(sl2->Strings[j].Trim().ToInt());
    }
  }
  __finally {
   	delete sl1;
   	delete sl2;
  }
	return il->Count;
}
//---------------------------------------------------------------------------
int __fastcall
StringsToIntList(TStrings *sl, TIntList *il, AnsiString ds, AnsiString cds) {
	il->Clear();
  TIntList *til = new TIntList;
  try {
  	for (int i = 0; i < sl->Count; ++i) {
     	if (StringToIntList(sl->Strings[i], til, ds, cds)) {
      	il->Assign(til, laOr);
				til->Clear();
      }
    }
  }
  __finally {
   	delete til;
  }
	return il->Count;
}
//---------------------------------------------------------------------------
void __fastcall IntListToStrings(TIntList *il, TStrings *sl) {
	sl->Clear();
  for (int i = 0; i < il->Count; ++i)
  	sl->Add(il->Items[i]);
}
//---------------------------------------------------------------------------
void __fastcall FillDefaults(TStrings *l, TStrings *dl) {
 	for (int i = 0; i < l->Count; ++i)
  	if (l->Strings[i].IsEmpty() && i < dl->Count)
			l->Strings[i] = dl->Strings[i];
}
//---------------------------------------------------------------------------
AnsiString __fastcall GetRegStr(TRegistry *reg, AnsiString key)
{
  AnsiString s = reg->ReadString(key);
  if (s.IsEmpty()) {
    AnsiString e = " \"";
    e += key;
    e += "\"    ";
    throw Exception(e);
  }
  return s;
}
//---------------------------------------------------------------------------
AnsiString __fastcall Encode(AnsiString s, AnsiString key)
{
	AnsiString rs;
  char c[2];
  c[1] = 0;
  for (int i = 0; i < s.Length(); ++i) {
  	c[0] = (s[i + 1] ^ key[i % key.Length() + 1]);
		if (!c[0])
    	c[0] = s[i + 1];
    rs += c;
  }
  return rs;
}
//---------------------------------------------------------------------------
AnsiString __fastcall ReadCrypt(TRegistry *reg, AnsiString name, AnsiString key)
{
	int bl = reg->GetDataSize(name);
  char *b = new char[++bl];
	AnsiString s;
	union {
  	int L;
    char C[sizeof(int)];
  } sl;
  try {
    b[reg->ReadBinaryData(name, b, bl)] = 0;
		for (int i = 0; i < 2 * sizeof(sl); ++i)
      if (i & 1)
      	sl.C[sizeof(sl) - (i >> 1) - 1] = b[i];
    s = AnsiString(b + 2 * sizeof(sl)).SubString(1, sl.L);
  }
  __finally {
   	delete[] b;
  }
  return Encode(s, key);
}
//---------------------------------------------------------------------------
void __fastcall WriteCrypt(TRegistry *reg, AnsiString name, AnsiString s, AnsiString key)
{
	s = Encode(s, key);
	union {
  	int L;
    char C[sizeof(int)];
  } sl;
  sl.L = s.Length();
  AnsiString ts = FormatDateTime("hhnnsszzz", Now());
  int bl = 2 * sizeof(sl) + sl.L + 1;
  if (bl < MinCryptLength)
  	bl = MinCryptLength;
  char *b = new char[bl];
  try {
  	int i;
		for (i = 0; i < 2 * sizeof(sl); ++i)
			b[i] = i & 1 ? sl.C[sizeof(sl) - (i >> 1) - 1] : ts[ts.Length() - (i >> 1)];
    randomize();
		for (int j = 1; i < bl; ++i, ++j)
    	b[i] = j <= sl.L ? s[j] : random(256);
    reg->WriteBinaryData(name, b, bl);
  }
  __finally {
  	delete[] b;
  }
}
//---------------------------------------------------------------------------
void __fastcall ReadCrypt(TRegistry *reg, AnsiString name, TStrings *sl, AnsiString key)
{
  sl->Text = ReadCrypt(reg, name, key);
}
//---------------------------------------------------------------------------
void __fastcall WriteCrypt(TRegistry *reg, AnsiString name, TStrings *sl, AnsiString key)
{
	WriteCrypt(reg, name, sl->Text, key);
}
//---------------------------------------------------------------------------
void __fastcall Read(TRegistry *reg, TStrings *sl)
{
	for (int i = 0; i < sl->Count; ++i)
  	sl->Values[sl->Names[i]] = reg->ReadString(sl->Names[i]);
}
//---------------------------------------------------------------------------
void __fastcall Write(TRegistry *reg, TStrings *sl)
{
	for (int i = 0; i < sl->Count; ++i)
  	reg->WriteString(sl->Names[i], sl->Values[sl->Names[i]]);
}
//---------------------------------------------------------------------------
AnsiString __fastcall ParamsToString(TParams *p, AnsiString sep)
{
  AnsiString s;
	for (int i = 0; i < p->Count; ++i) {
		if (i)
    	s += sep;
    s += p->Items[i]->Name + "=" + p->Items[i]->AsString;
  }
  return s;
}
//---------------------------------------------------------------------------
AnsiString __fastcall NormPath(AnsiString path)
{
  if (!path.IsEmpty() && path[path.Length()] != '\\')
	 	return path + "\\";
  else
  	return path;
}
//---------------------------------------------------------------------------
AnsiString __fastcall Increment(AnsiString s) {
  int p, l = s.Length();
  for (p = l; p; --p)
		if (!isdigit(s[p]))
      break;
	return s.SubString(1, p) + (s.SubString(p + 1, l - p).ToIntDef(0) + 1);
}
//---------------------------------------------------------------------------
/*
void StringToVarList(AnsiString s, AnsiString format, TList *l,
                     TList *defaults, AnsiString ds = " \t,") {
typedef enum {NO, UC, SC, US, SS, UI, SI, UL, SL, F, D, LD, SZ} TFormat;*/
/*
 * 	Recognized types (no case sensitivity):
 *	uc=cu				unsigned char		1	byte
 *	sc=cs=c			signed char			1	byte
 *	us=su	  		unsigned short	2	bytes
 *	ss=s				signed short		2	bytes
 *	ui=iu=u  		unsigned int		4	bytes
 *	si=is=i			signed int			4	bytes
 *	ul=lu	 			unsigned long		4	bytes
 *	sl=ls=l			signed long			4	bytes
 *	f		 		  	float 					4	bytes
 *  d=fl=lf			double  				8	bytes
 *	ld=dl	 			long double			10	bytes
 *  sz[]				string (char[] in "")	?	bytes	(last byte = 0)
 *  sz[N]				string (char[N] in "")	N	bytes (last byte = 0)
 */
//}
//---------------------------------------------------------------------------
AnsiString __fastcall SysTimeToString(const SYSTEMTIME &t)
{
  AnsiString s = t.wHour < 10 ? "0" : "";
  s += t.wHour;
  s += t.wMinute < 10 ? ":0" : ":";
  s += t.wMinute;
  s += t.wSecond < 10 ? ":0" : ":";
  s += t.wSecond;
  return s;
}
//---------------------------------------------------------------------------
AnsiString __fastcall SysDateToString(const SYSTEMTIME &t)
{
  AnsiString s = t.wDay < 10 ? "0" : "";
  s += t.wDay;
  s += t.wMonth < 10 ? ".0" : ".";
  s += t.wMonth;
  s += ".";
  s += t.wYear;
  return s;
}
//---------------------------------------------------------------------------
AnsiString __fastcall SysDateTimeToString(const SYSTEMTIME &t)
{
	return SysDateToString(t) + " " + SysTimeToString(t);
}
//---------------------------------------------------------------------------
bool __fastcall IsIPAddress(AnsiString s) {
  bool ret = true;
  TStringList *sl = new TStringList;
  try {
    if (SplitString(s, sl, ".") != 4)
      ret = false;
    else
      for (int i = 0; i < 4 && ret; ++i) {
        int v = sl->Strings[i].ToIntDef(-1);
        if (v < 0 || v > 255)
          ret = false;
      }
  }
  __finally {
    delete sl;
  }
  return ret;
}
//---------------------------------------------------------------------------
bool __fastcall IsHex(const AnsiString &s, int precision) {
  bool h = true;
  int l = s.Length();
  int r = 0;
  bool f = false;
  for (int i = 0; i < l && h; ++i) {
    char c = s.c_str()[i];
    if (c != ' ' || f) {
      f = true;
      if (c != '0' || r) {
        ++r;
        if (!(c >= '0' && c <= '9' ||
              c >= 'A' && c <= 'F' ||
              c >= 'a' && c <= 'f'))
          h = false;
      }
    }
  }
  if (precision && r > precision)
    h = false;
  return h;
}
//---------------------------------------------------------------------------
int __fastcall HexToInt(const AnsiString &s, int precision) {
  int l = s.Length();
  int v = 0;
  int r = 0;
  bool f = false;
  for (int i = 0; i < l; ++i) {
    char c = s.c_str()[i];
    if (c != ' ' || f) {
      f = true;
      if (c != '0' || r) {
        ++r;
        if (c >= '0' && c <= '9')
          v = (v << 4) | c - '0';
        else if (c >= 'A' && c <= 'F')
          v = (v << 4) | c - ('A' - 0xa);
        else if (c >= 'a' && c <= 'f')
          v = (v << 4) | c - ('a' - 0xa);
        else
          throw Exception("\"" + s + "\" is not valid hexadecimal value");
      }
    }
  }
  if (precision && r > precision) {
    throw Exception("\"" + s + "\" is not in range 0H..." +
                    AnsiString::StringOfChar('F', precision) + "H");
  }
  return v;
}
//---------------------------------------------------------------------------
AnsiString __fastcall MixString(const AnsiString &s1, const AnsiString &s2)
{
  int l1 = s1.Length();
  int l2 = s2.Length();
  int l = l1 > l2 ? l1 : l2;
  AnsiString s;
  for (int i = 1; i <= l; ++i) {
    s += i > l1 ? ' ' : s1[i];
    s += i > l2 ? ' ' : s2[i];
  }
  return s;
}
//---------------------------------------------------------------------------
AnsiString SubstrFrom(const AnsiString &s, const AnsiString &d)
{
  int i = s.Pos(d);
  if (i)
    return s.SubString(i, s.Length() + 1 - i);
  else
    return "";
}
//---------------------------------------------------------------------------
AnsiString SubstrAfter(const AnsiString &s, const AnsiString &d)
{
  int i = s.Pos(d);
  if (i && i + d.Length() < s.Length() + 1)
    return s.SubString(i + d.Length(), s.Length() + 1 - i - d.Length());
  else
    return "";
}
//---------------------------------------------------------------------------
AnsiString SubstrTo(const AnsiString &s, const AnsiString &d)
{
  int i = s.Pos(d);
  if (i)
    return s.SubString(1, i - 1 + d.Length());
  else
    return "";
}
//---------------------------------------------------------------------------
AnsiString SubstrBefore(const AnsiString &s, const AnsiString &d)
{
  int i = s.Pos(d);
  if (i)
    return s.SubString(1, i - 1);
  else
    return "";
}
//---------------------------------------------------------------------------
AnsiString SubstrFromTo(const AnsiString &s, const AnsiString &d1, const AnsiString &d2)
{
  return SubstrTo(SubstrFrom(s, d1), d2);
}
//---------------------------------------------------------------------------
AnsiString SubstrAfterBefore(const AnsiString &s, const AnsiString &d1, const AnsiString &d2)
{
	return SubstrBefore(SubstrAfter(s, d1), d2);
}
//---------------------------------------------------------------------------
AnsiString __fastcall GuidString()
{
	UUID uuid;
  memset(&uuid, 0, sizeof(uuid));
  UuidCreate(&uuid);
  unsigned char *b;
  UuidToString(&uuid, &b);
  AnsiString s((char*)(void*)b);
  RpcStringFree(&b);
  return s;
}
//---------------------------------------------------------------------------



