//---------------------------------------------------------------------------
/*
Copyright (c) 2013 Serge L. Ryadkow admin@opcgate.ru http://opcgate.ru 

Permission is hereby granted, free of charge, to any person obtaining a copy of this 
software and associated documentation files (the "Software"), to deal in the Software 
without restriction, including without limitation the rights to use, copy, modify, 
merge, publish, distribute, sublicense, and/or sell copies of the Software, and to 
permit persons to whom the Software is furnished to do so, subject to the following 
conditions:

The above copyright notice and this permission notice shall be included in all copies 
or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, 
INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR 
PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE 
FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR 
OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR 
OTHER DEALINGS IN THE SOFTWARE. 
*/
//---------------------------------------------------------------------------
//---------------------------------------------------------------------------
#include <vcl.h>

#pragma hdrstop

#include "UDPSock.h"

//---------------------------------------------------------------------------
#pragma package(smart_init)
//---------------------------------------------------------------------------
//---------------------------------------------------------------------------
//TWSA
//---------------------------------------------------------------------------
int TWSA::RefCnt = 0;
WSADATA TWSA::Data;
//---------------------------------------------------------------------------
TWSA::TWSA()
{
  if (!RefCnt++) {
    if (WSAStartup(WINSOCK_VERSION, &Data))
      throw Exception("WSAStartup() error");
  }
}
//---------------------------------------------------------------------------
TWSA::~TWSA()
{
  if (!--RefCnt)
    WSACleanup();
}
//---------------------------------------------------------------------------
static TWSA WSA;
//---------------------------------------------------------------------------
//---------------------------------------------------------------------------
//TSockAddr
//---------------------------------------------------------------------------
__fastcall TSockAddr::TSockAddr(const AnsiString &host, int port)
{
  sin_family = AF_INET;
  sin_port = htons(port);
  sin_addr.s_addr = inet_addr(host.c_str());
}
//---------------------------------------------------------------------------
__fastcall TSockAddr::TSockAddr(int port)
{
  sin_family = AF_INET;
  sin_port = htons(port);
  sin_addr.s_addr = INADDR_ANY;
}
//---------------------------------------------------------------------------
__fastcall TSockAddr::TSockAddr()
{
  sin_family = AF_INET;
  sin_port = 0;
  sin_addr.s_addr = INADDR_ANY;
}
//---------------------------------------------------------------------------
AnsiString __fastcall TSockAddr::Host2Str() const
{
  return inet_ntoa(sin_addr);
}
//---------------------------------------------------------------------------
AnsiString __fastcall TSockAddr::Addr2Str() const
{
  return Host2Str() + ":" + ntohs(sin_port);
}
//---------------------------------------------------------------------------
TSockAddr::operator bool()
{
  return sin_port;
}
//---------------------------------------------------------------------------
bool TSockAddr::operator !()
{
  return !sin_port;
}
//---------------------------------------------------------------------------
TSockAddr& TSockAddr::operator =(const TSockAddr &a)
{
  sin_family = a.sin_family;
  sin_port = a.sin_port;
  sin_addr.s_addr = a.sin_addr.s_addr;
  return *this;
}
//---------------------------------------------------------------------------
TSockAddr& TSockAddr::operator =(int port)
{
  sin_port = htons(port);
  return *this;
}
//---------------------------------------------------------------------------
bool operator == (const TSockAddr &a1, const TSockAddr &a2)
{
  return a1.sin_family == a2.sin_family && a1.sin_port == a2.sin_port &&
      a1.sin_addr.s_addr == a2.sin_addr.s_addr;
}
//---------------------------------------------------------------------------
bool operator != (const TSockAddr &a1, const TSockAddr &a2)
{
  return a1.sin_family != a2.sin_family || a1.sin_port != a2.sin_port ||
      a1.sin_addr.s_addr != a2.sin_addr.s_addr;
}
//---------------------------------------------------------------------------
//---------------------------------------------------------------------------
//TSocket
//---------------------------------------------------------------------------
TUDPSocket::TUDPSocket(int port)
{
  Socket = socket(AF_INET, SOCK_DGRAM, IPPROTO_UDP);
  if (Socket == INVALID_SOCKET)
    throw Exception("TUDPSocket::TUDPSocket():socket() error #" +
                    IntToHex(WSAGetLastError(), 4));
  TSockAddr a(port);
  if (bind(Socket, (SOCKADDR*)&a, sizeof(a)) == SOCKET_ERROR) {
    closesocket(Socket);
    throw Exception("TUDPSocket::TUDPSocket():bind() error #" +
                    IntToHex(WSAGetLastError(), 4));
  }
}
//---------------------------------------------------------------------------
TUDPSocket::TUDPSocket()
{
  Socket = socket(AF_INET, SOCK_DGRAM, IPPROTO_UDP);
  if (Socket == INVALID_SOCKET)
    throw Exception("TUDPSocket::TUDPSocket():socket() error #" +
                    IntToHex(WSAGetLastError(), 4));
  TSockAddr a;
  if (bind(Socket, (SOCKADDR*)&a, sizeof(a)) == SOCKET_ERROR) {
    closesocket(Socket);
    throw Exception("TUDPSocket::TUDPSocket():bind() error #" +
                    IntToHex(WSAGetLastError(), 4));
  }
}
//---------------------------------------------------------------------------
TUDPSocket::~TUDPSocket()
{
  closesocket(Socket);
}
//---------------------------------------------------------------------------
int TUDPSocket::RecvFrom(char *buff, int len, TSockAddr *addr)
{
  int as = sizeof(sockaddr_in);
  return recvfrom(Socket, buff, len, 0, (SOCKADDR*)addr, &as);
}
//---------------------------------------------------------------------------
int TUDPSocket::SendTo(char *buff, int len, TSockAddr *addr)
{
  return sendto(Socket, buff, len, 0, (SOCKADDR*)addr, sizeof(sockaddr_in));
}
//---------------------------------------------------------------------------

