//---------------------------------------------------------------------------
/*
Copyright (c) 2013 Serge L. Ryadkow admin@opcgate.ru http://opcgate.ru 

Permission is hereby granted, free of charge, to any person obtaining a copy of this 
software and associated documentation files (the "Software"), to deal in the Software 
without restriction, including without limitation the rights to use, copy, modify, 
merge, publish, distribute, sublicense, and/or sell copies of the Software, and to 
permit persons to whom the Software is furnished to do so, subject to the following 
conditions:

The above copyright notice and this permission notice shall be included in all copies 
or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, 
INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR 
PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE 
FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR 
OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR 
OTHER DEALINGS IN THE SOFTWARE. 
*/
//---------------------------------------------------------------------------
//---------------------------------------------------------------------------

#include <vcl.h>

#pragma hdrstop

#include <Registry.hpp>
#include <Math.h>

#include "Util.h"
#include "StringUtil.h"
//---------------------------------------------------------------------------
#pragma package(smart_init)
//---------------------------------------------------------------------------
class TBadD {
  public:
	  TBadD() : Lo(0x00000000), Hi(0xfff80000) {}
    operator double() const {return D;}
  private:
    union { struct { unsigned long Lo, Hi; }; double D; };
};
const double BadD = TBadD();
//---------------------------------------------------------------------------
double Round(const double &v) {
  return v < 0 ? (fmod(v, 1) < -0.5 ? floor(v) : ceil(v)) :
                 (fmod(v, 1) < 0.5 ? floor(v) : ceil(v));
}
//---------------------------------------------------------------------------
long double Round(const long double &v) {
  return v < 0 ? (fmodl(v, 1) < -0.5 ? floorl(v) : ceill(v)) :
                 (fmodl(v, 1) < 0.5 ? floorl(v) : ceill(v));
}
//---------------------------------------------------------------------------
TDateTime LastCycle(TDateTime cycle, TDateTime time) {
	return time - fmod(time, cycle);
}
//---------------------------------------------------------------------------
TDateTime NextCycle(TDateTime cycle, TDateTime time) {
	return LastCycle(cycle, time) + cycle;
}
//---------------------------------------------------------------------------
TDateTime NearCycle(TDateTime cycle, TDateTime time) {
	return LastCycle(cycle, time + double(cycle) / 2);
}
//---------------------------------------------------------------------------
AnsiString TimeToHex(TDateTime t)
{
  union {double t; unsigned char c[sizeof(double)];} v;
  v.t = t;
  AnsiString s;
  for (int i = 0; i < sizeof(double); ++i)
    s += IntToHex(v.c[i], 2);
  return s;
}
//---------------------------------------------------------------------------
TDateTime HexToTime(const AnsiString &s)
{
  union {double t; unsigned char c[sizeof(double)];} v;
  for (int i = 0; i < sizeof(double); ++i)
    v.c[i] = HexToInt(s.SubString(i * 2 + 1, 2));
  return v.t;
}
//---------------------------------------------------------------------------
TDateTime SecToTime(int s)
{
  return TDateTime(double(1) / 86400 * s);
}
//---------------------------------------------------------------------------
HANDLE CheckFileHandle(HANDLE h) {
 	if (h == INVALID_HANDLE_VALUE)
  	RaiseLastOSError();
  return h;
}
//---------------------------------------------------------------------------
TDateTime __fastcall MSToDateTime(int ms) {
  return TDateTime(double(ms) / 24 / 60 / 60 / 1000);
}
//---------------------------------------------------------------------------
TDateTime __fastcall SToDateTime(int s) {
  return TDateTime(double(s) / 24 / 60 / 60);
}
//---------------------------------------------------------------------------
//---------------------------------------------------------------------------
__fastcall TTimeout::TTimeout(TDateTime t) {
  Time = Now() + t;
}
//---------------------------------------------------------------------------
__fastcall TTimeout::TTimeout(int ms) {
  Time = Now() + MSToDateTime(ms);
}
//---------------------------------------------------------------------------
bool __fastcall TTimeout::Check() {
  return (Now() >= Time);
}
//---------------------------------------------------------------------------
AnsiString __fastcall GetRegStr(TRegistry *reg, AnsiString key)
{
  AnsiString s = reg->ReadString(key);
  if (s.IsEmpty()) {
    AnsiString e = " \"";
    e += key;
    e += "\"    ";
    throw Exception(e);
  }
  return s;
}
//---------------------------------------------------------------------------
void __fastcall DateTimeToFileTime(TDateTime dt, TFileTime &ft)
{
	TSystemTime st;
  TFileTime lft;
  DateTimeToSystemTime(dt, st);
  SystemTimeToFileTime(&st, &lft);
  LocalFileTimeToFileTime(&lft, &ft);
}
//---------------------------------------------------------------------------
void __fastcall FileTimeToDateTime(TDateTime &dt, const TFileTime &ft)
{
	TSystemTime st;
  TFileTime lft;
  FileTimeToLocalFileTime(&ft, &lft);
  FileTimeToSystemTime(&lft, &st);
  dt = SystemTimeToDateTime(st);
}
//---------------------------------------------------------------------------
TDateTime __fastcall FileTimeToDateTime(const TFileTime &ft)
{
	TSystemTime st;
  TFileTime lft;
  FileTimeToLocalFileTime(&ft, &lft);
  FileTimeToSystemTime(&lft, &st);
  return SystemTimeToDateTime(st);
}
//---------------------------------------------------------------------------
void __fastcall Pack(void *buff, void *data, int offset, int bits)
{
	unsigned char *wb = new unsigned char[offset % 8 ? (bits % 8 ? bits / 8 + 1 : bits / 8) :
                                                     (bits % 8 ? bits / 8 + 2 : bits / 8 + 1)];
  Pack(buff, data, offset, bits, wb);
  delete[] wb;
}
//---------------------------------------------------------------------------
void __fastcall UnPack(void *buff, void *data, int offset, int bits)
{
  int sb = bits % 8;
  int sB = bits / 8;
  int ob = offset % 8;
  int oB = offset / 8;
  int sBO = sB + (ob + sb) / 8 + ((ob + sb) % 8 ? 1 : 0);
	memcpy(data, ((unsigned char*)buff) + oB, sBO);
  if (ob)
	  ShiftR(data, ob, sBO);
  if (sb % 8) {
    unsigned char m = 0xff << (sb % 8);
    ((unsigned char*)data)[sB] &= ~m;
	} else if (sBO != sB)
    ((unsigned char*)data)[sB] = 0;
  if (sBO > sB + 1)
    ((unsigned char*)data)[sB + 1] = 0;
}
//---------------------------------------------------------------------------
void __fastcall Pack(void *buff, void *data, int offset, int bits, void *wb)
{
  int sb = bits % 8;
  int sB = bits / 8;
  int ob = offset % 8;
  int oB = offset / 8;
	memcpy(wb, data, sb ? sB + 1 : sB);
  ShiftL(wb, ob, sb + ob > 8 ? sB + 2 : sB + 1);
  unsigned char m = ~(sB ? 0x0 : 0xff << sb) << ob;
  ((unsigned char*)wb)[0] &= m;
  ((unsigned char*)buff)[oB] &= ~m;
  ((unsigned char*)buff)[oB] |= ((unsigned char*)wb)[0];
  if (sB + (ob + sb) / 8) {
  	memcpy(((unsigned char*)buff) + oB + 1, (unsigned char*)wb + 1, sB + (ob + sb) / 8 - 1);
	  if ((ob + sb) % 8) {
	    unsigned char m = 0xff << ((ob + sb) % 8);
  	  ((unsigned char*)wb)[sB + (ob + sb) / 8] &= ~m;
    	((unsigned char*)buff)[oB + sB + (ob + sb) / 8] &= m;
  	  ((unsigned char*)buff)[oB + sB + (ob + sb) / 8] |= ((unsigned char*)wb)[sB + (ob + sb) / 8];
    }
  }
}
//---------------------------------------------------------------------------
void __fastcall ShiftR(void *buff, int offset, int bytes)
{
  int ob = offset % 8;
  int oB = offset / 8;
  if (oB && bytes > oB) {
  	memmove(buff, ((unsigned char*)buff) + oB, bytes - oB);
    memset((unsigned char*)buff + bytes - oB, 0, oB);
  }
	if (ob) {
    for (int i = 0; i < bytes - 1; ++i)
      *((unsigned char*)buff + i) = *(unsigned short*)((unsigned char*)buff + i) >> ob;
    ((unsigned char*)buff)[bytes - 1] >>= ob;
  }
}
//---------------------------------------------------------------------------
void __fastcall ShiftL(void *buff, int offset, int bytes)
{
  int ob = offset % 8;
  int oB = offset / 8;
  if (oB && bytes > oB) {
  	memmove(((unsigned char*)buff) + oB, buff, bytes - oB);
    memset(buff, 0, oB);
  }
	if (ob) {
    for (int i = bytes - 2; i >= 0; --i)
      *((unsigned char*)buff + i + 1) = *(unsigned short*)((unsigned char*)buff + i) << ob >> 8;
    ((unsigned char*)buff)[0] <<= ob;
  }
}
//---------------------------------------------------------------------------
void __fastcall Enable(TControl *c, bool enable) {
	c->Enabled = enable;
  c->Refresh();
  TWinControl *wc = dynamic_cast<TWinControl*>(c);
  if (wc)
    for (int i = 0; i < wc->ControlCount; ++i)
      Enable(wc->Controls[i], enable);
}
//---------------------------------------------------------------------------
void __fastcall ExitWindowsEx(UINT flags) {
  HANDLE ht;
  TOKEN_PRIVILEGES tkp;
  Win32Check(OpenProcessToken(GetCurrentProcess(), TOKEN_ADJUST_PRIVILEGES | TOKEN_QUERY, &ht));
  Win32Check(LookupPrivilegeValue(NULL, SE_SHUTDOWN_NAME, &tkp.Privileges[0].Luid));
  tkp.PrivilegeCount = 1;  // one privilege to set
  tkp.Privileges[0].Attributes = SE_PRIVILEGE_ENABLED;
  Win32Check(AdjustTokenPrivileges(ht, FALSE, &tkp, 0, 0, 0));
  Win32Check(GetLastError() == ERROR_SUCCESS);
  Win32Check(ExitWindowsEx(flags, 0));
}
//---------------------------------------------------------------------------
TFileLock::TFileLock(THandleStream *s)
	:	Handle(reinterpret_cast<HANDLE>(s->Handle)),
  	From(0),
    Length(s->Size) {
	Win32Check(LockFile(Handle, From, 0, Length, 0));
}
//---------------------------------------------------------------------------
TFileLock::TFileLock(THandleStream *s, int attempts, int attemptTimeout)
	:	Handle(reinterpret_cast<HANDLE>(s->Handle)),
  	From(0),
    Length(s->Size) {
  while (!LockFile(Handle, From, 0, Length, 0))
  	if (--attempts)
    	Sleep(attemptTimeout);
    else
    	RaiseLastOSError();
}
//---------------------------------------------------------------------------
TFileLock::TFileLock(THandleStream *s, int attempts, int attemptTimeout,
              int from, int length)
	:	Handle(reinterpret_cast<HANDLE>(s->Handle)),
  	From(from),
    Length(length) {
  while (!LockFile(Handle, From, 0, Length, 0))
  	if (--attempts)
    	Sleep(attemptTimeout);
    else
    	RaiseLastOSError();
}
//---------------------------------------------------------------------------
TFileLock::~TFileLock() {
	Win32Check(UnlockFile(Handle, From, 0, Length, 0));
}
//---------------------------------------------------------------------------
TDateTime __fastcall GmtToLocal(const TDateTime &t) {
  SYSTEMTIME gst;
  SYSTEMTIME lst;
  FILETIME gft;
  FILETIME lft;
  DateTimeToSystemTime(t, gst);
  Win32Check(SystemTimeToFileTime(&gst, &gft));
  Win32Check(FileTimeToLocalFileTime(&gft, &lft));
  Win32Check(FileTimeToSystemTime(&lft, &lst));
  return SystemTimeToDateTime(lst);
}
//---------------------------------------------------------------------------
TDateTime __fastcall LocalToGmt(const TDateTime &t) {
  SYSTEMTIME lst;
  SYSTEMTIME gst;
  FILETIME lft;
  FILETIME gft;
  DateTimeToSystemTime(t, lst);
  Win32Check(SystemTimeToFileTime(&lst, &lft));
  Win32Check(LocalFileTimeToFileTime(&lft, &gft));
  Win32Check(FileTimeToSystemTime(&gft, &gst));
  return SystemTimeToDateTime(gst);
}
//---------------------------------------------------------------------------
int __fastcall GetIEVersion()
{
  int ret = 0;
  TRegistry *r = new TRegistry;
  try {
    r->RootKey = HKEY_LOCAL_MACHINE;
    if (!r->OpenKeyReadOnly("SOFTWARE\\Microsoft\\Internet Explorer"))
      throw Exception("Microsoft Internet Explorer not found");
    AnsiString s = r->ReadString("Version");
    if (s.IsEmpty())
      throw Exception("Microsoft Internet Explorer version not found");
    if (!isdigit(*s.c_str()))
      throw Exception("Unknown Microsoft Internet Explorer version");
    ret = *s.c_str() - '0';
  }
  __finally {
    delete r;
  }
  return ret;
}
//---------------------------------------------------------------------------





