#ifndef DataCallbackSinkH
#define DataCallbackSinkH

// DataCallbackSink.h : Declaration of the CDataCallbackSink

#include <vcl.h>

#include <opcda.h>
#include <opccomn.h>

#include "CustomSinks.h"

// It could be anything, even all numbers in 0:
// {F8FE7C40-F9C1-11d3-96DB-00902787286C}
DEFINE_GUID(CLSID_DataCallbackSink,
   0xf8fe7c40, 0xf9c1, 0x11d3, 0x96, 0xdb, 0x0, 0x90, 0x27, 0x87, 0x28, 0x6c);

// Event types:
typedef void __fastcall (__closure *TOnDataChangeEvent)(
      /* [in] */ DWORD dwTransid,
      /* [in] */ OPCHANDLE hGroup,
      /* [in] */ HRESULT hrMasterquality,
      /* [in] */ HRESULT hrMastererror,
      /* [in] */ DWORD dwCount,
      /* [size_is][in] */ OPCHANDLE __RPC_FAR *phClientItems,
      /* [size_is][in] */ VARIANT __RPC_FAR *pvValues,
      /* [size_is][in] */ WORD __RPC_FAR *pwQualities,
      /* [size_is][in] */ FILETIME __RPC_FAR *pftTimeStamps,
      /* [size_is][in] */ HRESULT __RPC_FAR *pErrors);

typedef void __fastcall (__closure *TOnReadCompleteEvent)(
      /* [in] */ DWORD dwTransid,
      /* [in] */ OPCHANDLE hGroup,
      /* [in] */ HRESULT hrMasterquality,
      /* [in] */ HRESULT hrMastererror,
      /* [in] */ DWORD dwCount,
      /* [size_is][in] */ OPCHANDLE __RPC_FAR *phClientItems,
      /* [size_is][in] */ VARIANT __RPC_FAR *pvValues,
      /* [size_is][in] */ WORD __RPC_FAR *pwQualities,
      /* [size_is][in] */ FILETIME __RPC_FAR *pftTimeStamps,
      /* [size_is][in] */ HRESULT __RPC_FAR *pErrors);

typedef void __fastcall (__closure *TOnWriteCompleteEvent)(
      /* [in] */ DWORD dwTransid,
      /* [in] */ OPCHANDLE hGroup,
      /* [in] */ HRESULT hrMastererr,
      /* [in] */ DWORD dwCount,
      /* [size_is][in] */ OPCHANDLE __RPC_FAR *pClienthandles,
      /* [size_is][in] */ HRESULT __RPC_FAR *pErrors);

typedef void __fastcall (__closure *TOnCancelCompleteEvent)(
      /* [in] */ DWORD dwTransid,
      /* [in] */ OPCHANDLE hGroup);

/////////////////////////////////////////////////////////////////////////////
// CDataCallbackSink
class ATL_NO_VTABLE CDataCallbackSink :
	public CComObjectRootEx<CComMultiThreadModel>,
	public CComCoClass<CDataCallbackSink, &CLSID_DataCallbackSink>,
  public IOPCDataCallback
{
// properties to create handlers in the form:
protected:
	TOnDataChangeEvent FEvDataChange;
	TOnReadCompleteEvent FEvReadComplete;
	TOnWriteCompleteEvent FEvWriteComplete;
	TOnCancelCompleteEvent FEvCancelComplete;

public:
	__property TOnDataChangeEvent EvDataChange = {read=FEvDataChange,
                                                 write=FEvDataChange};
	__property TOnReadCompleteEvent EvReadComplete = {read=FEvReadComplete,
                                                     write=FEvReadComplete};
	__property TOnWriteCompleteEvent EvWriteComplete = {read=FEvWriteComplete,
                                                       write=FEvWriteComplete};
	__property TOnCancelCompleteEvent EvCancelComplete = {read=FEvCancelComplete,
                                                         write=FEvCancelComplete};
public:
   CDataCallbackSink()
	{
	}

DECLARE_NOT_AGGREGATABLE(CDataCallbackSink)

DECLARE_PROTECT_FINAL_CONSTRUCT()

#pragma option push -vi-
BEGIN_COM_MAP(CDataCallbackSink)
	COM_INTERFACE_ENTRY(IOPCDataCallback)
END_COM_MAP()
#pragma option pop

// IOPCDataCallback: the sink implements this interface to receive the
// events from the OPC group.
public:
   virtual HRESULT STDMETHODCALLTYPE OnDataChange(
      /* [in] */ DWORD dwTransid,
      /* [in] */ OPCHANDLE hGroup,
      /* [in] */ HRESULT hrMasterquality,
      /* [in] */ HRESULT hrMastererror,
      /* [in] */ DWORD dwCount,
      /* [size_is][in] */ OPCHANDLE __RPC_FAR *phClientItems,
      /* [size_is][in] */ VARIANT __RPC_FAR *pvValues,
      /* [size_is][in] */ WORD __RPC_FAR *pwQualities,
      /* [size_is][in] */ FILETIME __RPC_FAR *pftTimeStamps,
      /* [size_is][in] */ HRESULT __RPC_FAR *pErrors);

   virtual HRESULT STDMETHODCALLTYPE OnReadComplete(
      /* [in] */ DWORD dwTransid,
      /* [in] */ OPCHANDLE hGroup,
      /* [in] */ HRESULT hrMasterquality,
      /* [in] */ HRESULT hrMastererror,
      /* [in] */ DWORD dwCount,
      /* [size_is][in] */ OPCHANDLE __RPC_FAR *phClientItems,
      /* [size_is][in] */ VARIANT __RPC_FAR *pvValues,
      /* [size_is][in] */ WORD __RPC_FAR *pwQualities,
      /* [size_is][in] */ FILETIME __RPC_FAR *pftTimeStamps,
      /* [size_is][in] */ HRESULT __RPC_FAR *pErrors);

   virtual HRESULT STDMETHODCALLTYPE OnWriteComplete(
      /* [in] */ DWORD dwTransid,
      /* [in] */ OPCHANDLE hGroup,
      /* [in] */ HRESULT hrMastererr,
      /* [in] */ DWORD dwCount,
      /* [size_is][in] */ OPCHANDLE __RPC_FAR *pClienthandles,
      /* [size_is][in] */ HRESULT __RPC_FAR *pErrors);

   virtual HRESULT STDMETHODCALLTYPE OnCancelComplete(
      /* [in] */ DWORD dwTransid,
      /* [in] */ OPCHANDLE hGroup);
};

// Instantiable sink declaration
typedef TCustomSink<CDataCallbackSink, &IID_IOPCDataCallback>
   CCreatableDataCallbackSink;

//////////////////////////////////////////
/// ShutdownCallback
//////////////////////////////////////////

DEFINE_GUID(CLSID_ShutdownCallbackSink,
   0xf8fe7c40, 0xf9c1, 0x11d3, 0x96, 0xdb, 0x0, 0x90, 0x27, 0x87, 0x28, 0x6c);

// Event types:
typedef void __fastcall (__closure *TOnShutdownEvent)(
      /* [in] */ LPCWSTR szReason);

/////////////////////////////////////////////////////////////////////////////
// CShutdownCallbackSink
class ATL_NO_VTABLE CShutdownCallbackSink :
	public CComObjectRootEx<CComMultiThreadModel>,
	public CComCoClass<CShutdownCallbackSink, &CLSID_ShutdownCallbackSink>,
  public IOPCShutdown
{
// properties to create handlers in the form:
protected:
	TOnShutdownEvent FEvShutdown;

public:
	__property TOnShutdownEvent EvShutdown = {read=FEvShutdown,
                                                 write=FEvShutdown};
public:
   CShutdownCallbackSink()
	{
	}

DECLARE_NOT_AGGREGATABLE(CShutdownCallbackSink)

DECLARE_PROTECT_FINAL_CONSTRUCT()

#pragma option push -vi-
BEGIN_COM_MAP(CShutdownCallbackSink)
	COM_INTERFACE_ENTRY(IOPCShutdown)
END_COM_MAP()
#pragma option pop

// IOPCServerCallback: the sink implements this interface to receive the
// events from the OPC server.
public:
   virtual HRESULT STDMETHODCALLTYPE ShutdownRequest(
      /* [in] */ LPCWSTR szReason);

};

// Instantiable sink declaration
typedef TCustomSink<CShutdownCallbackSink, &IID_IOPCShutdown>
   CCreatableShutdownCallbackSink;

#endif
