//---------------------------------------------------------------------------
/*
OPCGate - open source software, which is the OPC-Gateway, 
allowing the simplest way to add functionality of the OPC DA server 
to custom programs.

Copyright (C) 2013 Serge L. Ryadkow

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA

Contacts:
admin@opcgate.ru
http://opcgate.ru
*/
//---------------------------------------------------------------------------
//---------------------------------------------------------------------------
#include <vcl.h>

#pragma hdrstop

#include <IndyStatSSL.h>

#include "ClientU.h"
#include "ConfigU.h"
#include "LogFileU.h"
//---------------------------------------------------------------------------
#pragma package(smart_init)
//---------------------------------------------------------------------------
TClient *Client;
//---------------------------------------------------------------------------
__fastcall TClient::TClient()
{
	::LibInit();
  Socket = new TSocket(this);
}
//---------------------------------------------------------------------------
__fastcall TClient::~TClient()
{
  delete Socket;
}
//---------------------------------------------------------------------------
void __fastcall TClient::Connect()
{
	Socket->Connect();
}
//---------------------------------------------------------------------------
void __fastcall TClient::Disconnect()
{
  try {
		Socket->Disconnect();
  } catch(EIdConnClosedGracefully &e) {
  } catch(Exception &e) {
  	LogFile->Log(e, "TClient::Disconnect()");
  }
}
//---------------------------------------------------------------------------
bool __fastcall TClient::Connected()
{
	return Socket->Connected();
}
//---------------------------------------------------------------------------
void __fastcall TClient::Send(const AnsiString &s)
{
	Socket->WriteLn(s);
}
//---------------------------------------------------------------------------
AnsiString __fastcall TClient::Receive(int to)
{
  return Socket->ReadLn(EOL, to);
}
//---------------------------------------------------------------------------
void __fastcall TClient::OnConnected(TObject *Sender)
{
  if (LogFile->CheckFlag(lfClient))
		LogFile->Log("ClientConnect");
}
//---------------------------------------------------------------------------
void __fastcall TClient::OnDisconnected(TObject *Sender)
{
  if (LogFile->CheckFlag(lfClient))
		LogFile->Log("ClientDisconnect");
}
//---------------------------------------------------------------------------
void __fastcall TClient::OnStatus(TObject *Sender,
		const TIdStatus sts, const AnsiString msg)
{
  if (LogFile->CheckFlag(lfClient))
		LogFile->Log("ClientStatus: " + msg);
}
//---------------------------------------------------------------------------
//---------------------------------------------------------------------------
__fastcall TClient::TSocket::TSocket(TClient *owner)
	: TIdTCPClient(0)
{
  if (Config->GetSSL()) {
    SSL = new TConnectionInterceptOpenSSL(0);
    SSL->OnConnect = &SSLConnect;
    SSL->OnStatusInfo = &SSLStatusInfo;
    SSL->OnVerifyPeer = &SSLVerifyPeer;
    SSL->SSLOptions->CertFile = Config->GetModuleDir() + Config->GetCertFile();
    SSL->SSLOptions->KeyFile = Config->GetModuleDir() + Config->GetKeyFile();
    SSL->SSLOptions->Method = Sslopenssl::sslvSSLv23;
    SSL->SSLOptions->Mode = Sslopenssl::sslmClient;
    SSL->SSLOptions->VerifyMode = TSSLVerifyModeSet()
        << Sslopenssl::sslvrfPeer
        << Sslopenssl::sslvrfFailIfNoPeerCert
        << Sslopenssl::sslvrfClientOnce;
	} else
  	SSL = 0;
  OnConnected = &owner->OnConnected;
  OnDisconnected = &owner->OnDisconnected;
  OnStatus = &owner->OnStatus;
  Host = Config->GetHost();
  Port = Config->GetPort();
  if (SSL) {
		Intercept = SSL;
  	InterceptEnabled = true;
  }
}
//---------------------------------------------------------------------------
__fastcall TClient::TSocket::~TSocket()
{
	if (SSL) {
	  InterceptEnabled = false;
  	Intercept = 0;
	  delete SSL;
  }
}
//---------------------------------------------------------------------------
void __fastcall TClient::TSocket::SSLConnect(TObject *Sender)
{
  if (LogFile->CheckFlag(lfSSL))
		LogFile->Log("SSL Connect");
}
//---------------------------------------------------------------------------
void __fastcall TClient::TSocket::SSLStatusInfo(AnsiString msg)
{
  if (LogFile->CheckFlag(lfSSL))
		LogFile->Log("SSL Status: " + msg);
}
//---------------------------------------------------------------------------
bool __fastcall TClient::TSocket::SSLVerifyPeer(TX509 *cert, int ok)
{
  if (LogFile->CheckFlag(lfSSL))
		LogFile->Log("SSL Verify Peer: " + cert->Subject->OneLine + ", ok=" + ok);
	return true;
}
//---------------------------------------------------------------------------

