//---------------------------------------------------------------------------
/*
OPCGate - open source software, which is the OPC-Gateway, 
allowing the simplest way to add functionality of the OPC DA server 
to custom programs.

Copyright (C) 2013 Serge L. Ryadkow

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA

Contacts:
admin@opcgate.ru
http://opcgate.ru
*/
//---------------------------------------------------------------------------
//---------------------------------------------------------------------------
#include <vcl.h>

#pragma hdrstop

#include <fstream.h>

#include "CmdParsU.h"
//---------------------------------------------------------------------------
#pragma package(smart_init)
//---------------------------------------------------------------------------
TCmdLineParser *CmdLineParser;
//---------------------------------------------------------------------------
__fastcall TCmdLineParser::TCmdLineParser(int argc, char* argv[])
{
  OS = 0;
  F = RegF = UnRegF = GuidF = TagsF = false;
	try {
    for (int i = 1; i < argc; ++i) {
      if (!strcmp(argv[i], "-o") || !strcmp(argv[i], "--out")) {
      	if (++i < argc)
        	OS = new ofstream(argv[i], ios::app);
        else
        	throw Exception("No output file specified");
        F = true;
      } else if (!strcmp(argv[i], "-h") || !strcmp(argv[i], "--help")) {
      	Help();
        F = true;
      } else if (!strcmp(argv[i], "-u") || !strcmp(argv[i], "--unregister"))
      	F = UnRegF = true;
      else if (!strcmp(argv[i], "-g") || !strcmp(argv[i], "--guidgen"))
      	F = GuidF = true;
      else if (!strcmp(argv[i], "-r") || !strcmp(argv[i], "--register"))
      	F = RegF = true;
      else if (!strcmp(argv[i], "-t") || !strcmp(argv[i], "--tags"))
      	F = TagsF = true;
    }
  } catch(Exception &e) {
  	cerr << '\n' << "Command line parser error:\n" << e.Message.c_str() << endl;
    Help();
  }
}
//---------------------------------------------------------------------------
__fastcall TCmdLineParser::~TCmdLineParser()
{
	delete OS;
}
//---------------------------------------------------------------------------
void __fastcall TCmdLineParser::Help()
{
	GetOS() << "\n\nCommand line options\n"
  	<< "-o filename | --out filename - redirect output to file\n"
    << "-h | --help - show this help screen\n"
    << "-u | --unregister - unregister OPC server\n"
    << "-g | --guidgen - generate new GUID\n"
    << "-r | --register - register OPC server\n"
    << "-t | --tags - refresh tags list\n"
    << endl;
}
//---------------------------------------------------------------------------

