//---------------------------------------------------------------------------
/*
OPCGate - open source software, which is the OPC-Gateway, 
allowing the simplest way to add functionality of the OPC DA server 
to custom programs.

Copyright (C) 2013 Serge L. Ryadkow

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA

Contacts:
admin@opcgate.ru
http://opcgate.ru
*/
//---------------------------------------------------------------------------
//---------------------------------------------------------------------------
#include <vcl.h>

#pragma hdrstop

#include <constream.h>

#include "LogFileU.h"
#include "ConfigU.h"
//---------------------------------------------------------------------------
#pragma package(smart_init)
//---------------------------------------------------------------------------
TLogFile *LogFile;
//---------------------------------------------------------------------------
__fastcall TLogFile::TLogFile()
{
  CS = new TCriticalSection();
  Flags = TLogFlag(Config->GetLogFlags());
	if (Flags) {
	  LogFile = new ofstream(ChangeFileExt(__argv[0], ".log").c_str(),
  	                       ios::out | ios::app);
    *LogFile << endl;
	} else
  	LogFile = 0;
  ErrFile = new ofstream(ChangeFileExt(__argv[0], ".err").c_str(),
                         ios::out | ios::app);
}
//---------------------------------------------------------------------------
__fastcall TLogFile::~TLogFile()
{
  delete LogFile;
  delete ErrFile;
  delete CS;
}
//---------------------------------------------------------------------------
void __fastcall TLogFile::Log(const AnsiString &s)
{
  AnsiString ls = FormatDateTime(
      "dd\".\"mm\".\"yyyy\" \"hh\":\"nn\":\"ss\".\"zzz", Now()) +  " " + s;
  CS->Enter();
  try {
    *LogFile << ls.c_str() << endl;
  } __finally {
    CS->Leave();
  }
}
//---------------------------------------------------------------------------
void __fastcall TLogFile::Log(Exception &e, const AnsiString &info)
{
  AnsiString ls = FormatDateTime(
		  "dd\".\"mm\".\"yyyy\" \"hh\":\"nn\":\"ss\".\"zzz", Now()) +  " " +
  				"Error " + info + " Class: " + e.ClassName() +
      		" File: " + __throwFileName + " Line: " + __throwLineNumber +
      		" Message: " + e.Message;
  CS->Enter();
  try {
    *ErrFile << ls.c_str() << endl;
    if (LogFile && (Flags & lfError))
	    *LogFile << ls.c_str() << endl;
  } __finally {
    CS->Leave();
  }
}
//---------------------------------------------------------------------------

