//---------------------------------------------------------------------------
/*
OPCGate - open source software, which is the OPC-Gateway, 
allowing the simplest way to add functionality of the OPC DA server 
to custom programs.

Copyright (C) 2013 Serge L. Ryadkow

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA

Contacts:
admin@opcgate.ru
http://opcgate.ru
*/
//---------------------------------------------------------------------------
//---------------------------------------------------------------------------
#include <vcl.h>

#pragma hdrstop

#include <constream.h>
#include <process.h>

#include <IPC.h>
#include <OPCSrvr.h>
#include <StringUtil.h>

#include "CmdParsU.h"
#include "ConfigU.h"
#include "LogFileU.h"
#include "ClientU.h"
#include "TagsU.h"
#include "ServerU.h"
#include "ProtocolU.h"
//---------------------------------------------------------------------------
BOOL __stdcall CtrlBreakH(DWORD dwCtrlType);
IPC::TEvent CtrlBreakE;
void __fastcall SetLocale();
//---------------------------------------------------------------------------
#pragma argsused
int main(int argc, char* argv[])
{
  int ret = 0;
  SetLocale();
  unsigned int pid = getpid();
  AnsiString pidStr = pid;
  try {
  	cout << "Parsing command line... " << flush;
    CmdLineParser = new TCmdLineParser(argc, argv);
  	cout << "OK" << endl;
    try {

      CmdLineParser->GetOS() << "Starting... PID=" << pid << endl;

      CmdLineParser->GetOS() << "Create Config... " << flush;
      Config = new TConfig();
      CmdLineParser->GetOS() << "OK" << endl;
      try {
        CmdLineParser->GetOS() << "Create LogFile... " << flush;
        LogFile = new TLogFile();
        CmdLineParser->GetOS() << "OK" << endl;
        try {
          try {
            TServer::Configure();
            if (CmdLineParser->GetUnRegF()) {
              CmdLineParser->GetOS() << "\n\nUnregistering OPC server... "
                  << flush;
              TOPCSrvr::UnRegister();
              CmdLineParser->GetOS() << "OK\n" << endl;
            }
            if (CmdLineParser->GetGuidF()) {
              AnsiString guid = GuidString();
              CmdLineParser->GetOS() << "\n\nNew generated GUID: "
                  << guid.c_str() << '\n' << endl;
            }
            if (CmdLineParser->GetRegF()) {
              CmdLineParser->GetOS() << "\n\nRegistering OPC server... "
                  << flush;
              TOPCSrvr::Register();
              CmdLineParser->GetOS() << "OK\n" << endl;
            }
            if (!CmdLineParser->GetF() || CmdLineParser->GetTagsF()) {

              if (LogFile->CheckFlag(lfStartStop))
                LogFile->Log("Starting... PID=" + pidStr);

              if (LogFile->CheckFlag(lfCfg))
                LogFile->Log(Config->ToString());

              CmdLineParser->GetOS() << "Create Client... " << flush;
              Client = new TClient();
              CmdLineParser->GetOS() << "OK" << endl;
              try {
                CmdLineParser->GetOS() << "Create Protocol... " << flush;
                Protocol = new TProtocol();
                CmdLineParser->GetOS() << "OK" << endl;
                try {
                  CmdLineParser->GetOS() << "Create Tags... " << flush;
                  Tags = new TTags();
                  CmdLineParser->GetOS() << "OK" << endl;
                  try {
                    if (!CmdLineParser->GetF()) {
                      CmdLineParser->GetOS() << "Create Server... " << flush;
                      Server = new TServer();
                      CmdLineParser->GetOS() << "OK" << endl;
                      try {
                        CmdLineParser->GetOS() << "Starting Server... "
                            << flush;
                        Server->Start();
                        CmdLineParser->GetOS() << "OK" << endl;
                        try {
                          CmdLineParser->GetOS() << "Starting Protocol... "
                              << flush;
                          Protocol->Start();
                          CmdLineParser->GetOS() << "OK" << endl;
                          try {

                            Win32Check(SetConsoleCtrlHandler(&CtrlBreakH, true));

                            CmdLineParser->GetOS() << "Started OK PID=" << pid
                                << endl;
                            if (LogFile->CheckFlag(lfStartStop))
                              LogFile->Log("Started OK PID=" + pidStr);
                            CtrlBreakE.Wait();
                            if (LogFile->CheckFlag(lfStartStop))
                              LogFile->Log("Stopping... PID=" + pidStr);
                            CmdLineParser->GetOS() << "Stopping... PID=" << pid
                                << endl;

                          } __finally {
                            CmdLineParser->GetOS() << "Stopping Protocol... "
                                << flush;
                            Protocol->Stop();
                            CmdLineParser->GetOS() << "OK" << endl;
                          }
                        } __finally {
                          CmdLineParser->GetOS() << "Stopping Server... "
                              << flush;
                          Server->Stop();
                          CmdLineParser->GetOS() << "OK" << endl;
                        }
                      } __finally {
                        CmdLineParser->GetOS() << "Destroy Server... " << flush;
                        delete Server;
                        CmdLineParser->GetOS() << "OK" << endl;
                      }
                    }
                  } __finally {
                    CmdLineParser->GetOS() << "Destroy Tags... " << flush;
                    delete Tags;
                    CmdLineParser->GetOS() << "OK" << endl;
                  }
                } __finally {
                  CmdLineParser->GetOS() << "Destroy Protocol... " << flush;
                  delete Protocol;
                  CmdLineParser->GetOS() << "OK" << endl;
                }
              } __finally {
                CmdLineParser->GetOS() << "Destroy Client... " << flush;
                delete Client;
                CmdLineParser->GetOS() << "OK" << endl;
              }

              if (LogFile->CheckFlag(lfStartStop))
                LogFile->Log("Stopped OK PID=" + pidStr);

            }
          } catch(Exception &e) {
            LogFile->Log(e, "main()");
            throw;
          }
        } __finally {
          CmdLineParser->GetOS() << "Destroy LogFile... " << flush;
          delete LogFile;
          CmdLineParser->GetOS() << "OK" << endl;
        }
      } __finally {
        CmdLineParser->GetOS() << "Destroy Config... " << flush;
        delete Config;
        CmdLineParser->GetOS() << "OK" << endl;
      }

      CmdLineParser->GetOS() << "Stopped OK PID=" << pid << endl;

    } __finally {
      delete CmdLineParser;
    }
  } catch(Exception &e) {
	  AnsiString ls = FormatDateTime(
			  "dd\".\"mm\".\"yyyy\" \"hh\":\"nn\":\"ss\".\"zzz", Now()) +
  					" Error main() Class: " + e.ClassName() +
      			" File: " + __throwFileName + " Line: " + __throwLineNumber +
      			" Message: " + e.Message;
    cerr << ls.c_str() << endl;
    ret = 1;
  }
  return ret;
}
//---------------------------------------------------------------------------
//---------------------------------------------------------------------------
bool CtrlBreakF = false;
//---------------------------------------------------------------------------
BOOL __stdcall CtrlBreakH(DWORD dwCtrlType)
{
  if (dwCtrlType == CTRL_BREAK_EVENT && !CtrlBreakF) {
    CtrlBreakF = true;
    CtrlBreakE.Set();
    return true;
  } else
    return false;
}
//---------------------------------------------------------------------------
void __fastcall SetLocale()
{
	/*char*/ DecimalSeparator = '.';
  /*char*/ DateSeparator = '.';
  /*AnsiString*/ ShortDateFormat = "dd.MM.yy";
  /*AnsiString*/ LongDateFormat = "d MMMM yyyy .";
  /*char*/ TimeSeparator = ':';
  /*AnsiString*/ TimeAMString = "";
  /*AnsiString*/ TimePMString = "";
  /*AnsiString*/ ShortTimeFormat = "h:mm";
  /*AnsiString*/ LongTimeFormat = "h:mm:ss";                                   
}
//---------------------------------------------------------------------------


