//---------------------------------------------------------------------------
/*
OPCGate - open source software, which is the OPC-Gateway, 
allowing the simplest way to add functionality of the OPC DA server 
to custom programs.

Copyright (C) 2013 Serge L. Ryadkow

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA

Contacts:
admin@opcgate.ru
http://opcgate.ru
*/
//---------------------------------------------------------------------------
//---------------------------------------------------------------------------
#include <vcl.h>

#pragma hdrstop

#include <DateUtils.hpp>

#include "ProtocolU.h"
#include "ClientU.h"
#include "LogFileU.h"
#include "ConfigU.h"
#include "TagsU.h"
#include "ServerU.h"
//---------------------------------------------------------------------------
#pragma package(smart_init)
//---------------------------------------------------------------------------
TProtocol *Protocol;
//---------------------------------------------------------------------------
__fastcall TProtocol::TProtocol()
{
}
//---------------------------------------------------------------------------
__fastcall TProtocol::~TProtocol()
{
}
//---------------------------------------------------------------------------
void __fastcall TProtocol::Start()
{
	Thread = new TThrd(this);
}
//---------------------------------------------------------------------------
void __fastcall TProtocol::Stop()
{
	delete Thread;
}
//---------------------------------------------------------------------------
void __fastcall TProtocol::GetTags(TStrings *sl)
{
  if (LogFile->CheckFlag(lfProtocol))
	 	LogFile->Log("TProtocol::GetTags()");
  if (!Client->Connected())
  	Client->Connect();
	Client->Send("TAGS");
  AnsiString res = Client->Receive(Config->GetProtocolTO());
	int l = res.Length();
	if (!l)
  	throw Exception("TProtocol::GetTags() timeout");
	if (l < 6)
  	throw Exception("TProtocol::GetTags() invalid response: " + res);
  if (res.SubString(1, 4) != "TAGS")
  	throw Exception("TProtocol::GetTags() invalid response: " + res);
  int n = res.SubString(6, l - 5).ToInt();
  for (int i = 0; i < n; ++i) {
  	AnsiString res = Client->Receive(Config->GetProtocolTO());
		int l = res.Length();
		if (!l)
  		throw Exception("TProtocol::GetTags() timeout");
    if (l < 5)
      throw Exception("TProtocol::GetTags() invalid response: " + res);
    if (res.SubString(1, 3) != "TAG")
      throw Exception("TProtocol::GetTags() invalid response: " + res);
    sl->Add(res.SubString(5, l - 4));
  }
  if (LogFile->CheckFlag(lfProtocol))
 		LogFile->Log("TProtocol::GetTags(): " + AnsiString(n) + " tags");
}
//---------------------------------------------------------------------------
AnsiString __fastcall TProtocol::GetHash()
{
  if (LogFile->CheckFlag(lfProtocol))
	 	LogFile->Log("TProtocol::GetHash()");
	Client->Send("HASH");
  AnsiString res = Client->Receive(Config->GetProtocolTO());
	int l = res.Length();
	if (!l)
  	throw Exception("TProtocol::GetHash() timeout");
	if (l < 6)
  	throw Exception("TProtocol::GetHash() invalid response: " + res);
  if (res.SubString(1, 4) != "HASH")
  	throw Exception("TProtocol::GetHash() invalid response: " + res);
  AnsiString hash = res.SubString(6, l - 5);
  if (LogFile->CheckFlag(lfProtocol))
 		LogFile->Log("TProtocol::GetHash(): " + hash);
  return hash;
}
//---------------------------------------------------------------------------
void __fastcall TProtocol::GetValues(TStrings *sl)
{
  if (LogFile->CheckFlag(lfProtocol))
	 	LogFile->Log("TProtocol::GetValues()");
	Client->Send("VALUES");
  AnsiString res = Client->Receive(Config->GetProtocolTO());
	int l = res.Length();
	if (!l)
  	throw Exception("TProtocol::GetValues() timeout");
	if (l < 8)
  	throw Exception("TProtocol::GetValues() invalid response: " + res);
  if (res.SubString(1, 6) != "VALUES")
  	throw Exception("TProtocol::GetValues() invalid response: " + res);
  int n = res.SubString(8, l - 7).ToInt();
  for (int i = 0; i < n; ++i) {
  	AnsiString res = Client->Receive(Config->GetProtocolTO());
		int l = res.Length();
		if (!l)
  		throw Exception("TProtocol::GetValues() timeout");
    if (l < 7)
      throw Exception("TProtocol::GetValues() invalid response: " + res);
    if (res.SubString(1, 5) != "VALUE")
      throw Exception("TProtocol::GetValues() invalid response: " + res);
    sl->Add(res.SubString(7, l - 6));
  }
  if (LogFile->CheckFlag(lfProtocol))
	 	LogFile->Log("TProtocol::GetValues(): " + AnsiString(n) + " values");
}
//---------------------------------------------------------------------------
int __fastcall TProtocol::GetSubsValues(TStrings *sl)
{
  int n = 0;
  AnsiString res = Client->Receive(Config->GetProtocolTO());
	int l = res.Length();
  if (l >= 8)
	  if (res.SubString(1, 6) == "VALUES") {
		  n = res.SubString(8, l - 7).ToInt();
      for (int i = 0; i < n; ++i) {
        AnsiString res = Client->Receive(Config->GetProtocolTO());
        int l = res.Length();
        if (!l)
          throw Exception("TProtocol::GetSubsValues() timeout");
        if (l < 7)
          throw Exception("TProtocol::GetSubsValues() invalid response: " + res);
        if (res.SubString(1, 5) != "VALUE")
          throw Exception("TProtocol::GetSubsValues() invalid response: " + res);
        sl->Add(res.SubString(7, l - 6));
      }
    }
  if (n && LogFile->CheckFlag(lfProtocol))
	 	LogFile->Log("TProtocol::GetSubsValues(): " + AnsiString(n) + " values");
  return n;
}
//---------------------------------------------------------------------------
void __fastcall TProtocol::Subscribe()
{
  if (LogFile->CheckFlag(lfProtocol))
	 	LogFile->Log("TProtocol::Subscribe()");
	Client->Send("SUBSCRIBE");
  AnsiString res = Client->Receive(Config->GetProtocolTO());
	int l = res.Length();
	if (!l)
  	throw Exception("TProtocol::Subscribe() timeout");
	if (res != "SUBSCRIBE_ACK")
  	throw Exception("TProtocol::Subscribe() invalid response: " + res);
	Client->Send("VALUES");///????????????????????
  if (LogFile->CheckFlag(lfProtocol))
	 	LogFile->Log("TProtocol::Subscribe(): ACK");
}
//---------------------------------------------------------------------------
void __fastcall TProtocol::SetConnTO()
{
  if (LogFile->CheckFlag(lfProtocol))
	 	LogFile->Log("TProtocol::SetConnTO()");
	Client->Send("PAR_CONNTO " + AnsiString(Config->GetProtocolSrvConnTO()));
  AnsiString res = Client->Receive(Config->GetProtocolTO());
	int l = res.Length();
	if (!l)
  	throw Exception("TProtocol::SetConnTO() timeout");
	if (res != "PAR_CONNTO_ACK")
  	throw Exception("TProtocol::SetConnTO() invalid response: " + res);
  if (LogFile->CheckFlag(lfProtocol))
	 	LogFile->Log("TProtocol::SetConnTO(): ACK");
}
//---------------------------------------------------------------------------
void __fastcall TProtocol::SetSubsTO()
{
  if (LogFile->CheckFlag(lfProtocol))
	 	LogFile->Log("TProtocol::SetSubsTO()");
	Client->Send("PAR_SUBSTO " + AnsiString(Config->GetProtocolSrvSubsTO()));
  AnsiString res = Client->Receive(Config->GetProtocolTO());
	int l = res.Length();
	if (!l)
  	throw Exception("TProtocol::SetSubsTO() timeout");
	if (res != "PAR_SUBSTO_ACK")
  	throw Exception("TProtocol::SetSubsTO() invalid response: " + res);
  if (LogFile->CheckFlag(lfProtocol))
	 	LogFile->Log("TProtocol::SetSubsTO(): ACK");
}
//---------------------------------------------------------------------------
void __fastcall TProtocol::Ping()
{
  if (LogFile->CheckFlag(lfProtocol))
	 	LogFile->Log("TProtocol::Ping()");
	Client->Send("PING");
}
//---------------------------------------------------------------------------
//---------------------------------------------------------------------------
__fastcall TProtocol::TThrd::TThrd(TProtocol *prot)
	:	TThread(true)
{
	Prot = prot;
  Cycle = Config->GetProtocolCycle();
  Sts = sInit;
  Resume();
}
//---------------------------------------------------------------------------
__fastcall TProtocol::TThrd::~TThrd()
{
	Terminate();
  WaitFor();
}
//---------------------------------------------------------------------------
//virtual
void __fastcall TProtocol::TThrd::Execute()
{
  TStringList *sl;
	while (!Terminated) {
    bool err = false;
		try {
      switch (Sts) {
      	case sInit:
      	case sDisconn:
		    	if (!Client->Connected())
    		  	Client->Connect();
					Prot->SetConnTO();
          if (Config->GetProtocolType() == 2) //Subscr
						Prot->SetSubsTO();
          Server->SetSTConnections(true);
          if (Prot->GetHash() == Tags->GetHash())
            Server->SetSTTagsHashGood(true);
          else {
            Server->SetSTTagsHashGood(false);
            Sts = sBadHash;
            throw Exception("Bad tags hash");
          }
          if (Config->GetProtocolType() == 2) {//Subscr
            Prot->Subscribe();
            PingTime = Now();
            Sts = sSubscr;
          } else {
            sl = new TStringList();
            try {
              Prot->GetValues(sl);
              Server->SetValues(sl);
            } __finally {
              delete sl;
            }
            Sts = sConn;
          }
        	break;
        case sConn:
          sl = new TStringList();
          try {
          	Prot->GetValues(sl);
            Server->SetValues(sl);
          } __finally {
          	delete sl;
          }
        	break;
        case sSubscr:
          TDateTime now = Now();
          if (::MilliSecondsBetween(now, PingTime) > Cycle) {
            Prot->Ping();
            PingTime = now;
          } else {
            sl = new TStringList();
            try {
              if (Prot->GetSubsValues(sl))
                Server->SetValues(sl);
            } __finally {
              delete sl;
            }
          }
        	break;
      }
    } catch(Exception &e) {
    	LogFile->Log(e, "TProtocol::TThrd::Execute()");
      err = true;
    }
    if (err) {
    	if (Client->Connected())
      	Client->Disconnect();
      Sts = sDisconn;
      Server->SetSTConnections(false);
      Server->SetBad();
    }
    if (Sts != sSubscr) {
      TDateTime t0 = Now();
      while (::MilliSecondsBetween(t0, Now()) < Cycle)
		   	::Sleep(100);
    }
  }
}
//---------------------------------------------------------------------------

