//---------------------------------------------------------------------------
/*
OPCGate - open source software, which is the OPC-Gateway, 
allowing the simplest way to add functionality of the OPC DA server 
to custom programs.

Copyright (C) 2013 Serge L. Ryadkow

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA

Contacts:
admin@opcgate.ru
http://opcgate.ru
*/
//---------------------------------------------------------------------------
//---------------------------------------------------------------------------
#include <vcl.h>

#pragma hdrstop

#include <dir.h>

#include <OPCSrvr.h>
#include <StringUtil.h>
#include <Util.h>

#include "ServerU.h"
#include "ConfigU.h"
#include "LogFileU.h"
#include "TagsU.h"
#include "CmdParsU.h"
//---------------------------------------------------------------------------
#pragma package(smart_init)
//---------------------------------------------------------------------------
TServer *Server;
//---------------------------------------------------------------------------
const char OPCClsID[MAXPATH] = "LightOPC Sample server 1";
const char OPCProgID[MAXPATH] = "OPC.LightOPC.1";
const loVendorInfo OPCVendor = {
  1, //Major
  0, //Minor
  1, //Build
  0, //Reserv
  "OPCGate OPC Server"
};
// {BAF849C4-9DBC-49f5-9ACE-DD73648BC8AD}
const GUID OPCGUID =
{ 0xbaf849c4, 0x9dbc, 0x49f5, { 0x9a, 0xce, 0xdd, 0x73, 0x64, 0x8b, 0xc8, 0xad } };
//---------------------------------------------------------------------------
__fastcall TServer::TServer()
{
  N = Tags->GetCount();
  OPC = new TOPCSrvr(1 + N + stSysTagsCount, Config->GetOPCRefRate(),
  		Config->GetOPCFlags());
  OPC->AddBool(stConnection, "OPCGate.Connection");
  OPC->AddBool(stTagsHashGood, "OPCGate.TagsHashGood");
  ///////////////
  Types = new TType[N];
  TStringList *sl = new TStringList();
  try {
	  for (int i = 0; i < N; ++i) {
    	if (::SplitString(Tags->GetList()->Strings[i], sl, " ") != 4)
      	throw Exception("TServer::TServer() invalid tag format: " +
        		Tags->GetList()->Strings[i]);
      int id = sl->Strings[0].ToIntDef(0);
      if (id < 1)
      	throw Exception("TServer::TServer() invalid tag format: " +
        		Tags->GetList()->Strings[i]);
      if (id > N)
      	throw Exception("TServer::TServer() invalid tag index: " +
        		Tags->GetList()->Strings[i]);
      if (sl->Strings[2] == "b")
        Types[id - 1] = tB;
      else if (sl->Strings[2] == "i1")
        Types[id - 1] = tI1;
      else if (sl->Strings[2] == "u1")
        Types[id - 1] = tU1;
      else if (sl->Strings[2] == "i2")
        Types[id - 1] = tI2;
      else if (sl->Strings[2] == "u2")
        Types[id - 1] = tU2;
      else if (sl->Strings[2] == "i4")
        Types[id - 1] = tI4;
      else if (sl->Strings[2] == "u4")
        Types[id - 1] = tU4;
      else if (sl->Strings[2] == "r4")
        Types[id - 1] = tR4;
      else if (sl->Strings[2] == "r8")
        Types[id - 1] = tR8;
      else
      	throw Exception("TServer::TServer() invalid tag format: " +
        		Tags->GetList()->Strings[i]);
      int acc;
      if (sl->Strings[3] == "r")
      	acc = OPC_READABLE;
      else if (sl->Strings[3] == "rw")
      	acc = OPC_READABLE | OPC_WRITEABLE;
      else
      	throw Exception("TServer::TServer() invalid tag format: " +
        		Tags->GetList()->Strings[i]);
      switch (Types[id - 1]) {
      	case tB:
				  OPC->AddBool(id + stSysTagsCount, sl->Strings[1].c_str(), acc);
        	break;
      	case tI1:
				  OPC->AddByte(id + stSysTagsCount, sl->Strings[1].c_str(), acc);
        	break;
      	case tU1:
				  OPC->AddUByte(id + stSysTagsCount, sl->Strings[1].c_str(), acc);
        	break;
      	case tI2:
				  OPC->AddShort(id + stSysTagsCount, sl->Strings[1].c_str(), acc);
        	break;
      	case tU2:
				  OPC->AddUShort(id + stSysTagsCount, sl->Strings[1].c_str(), acc);
        	break;
      	case tI4:
				  OPC->AddLong(id + stSysTagsCount, sl->Strings[1].c_str(), acc);
        	break;
      	case tU4:
				  OPC->AddULong(id + stSysTagsCount, sl->Strings[1].c_str(), acc);
        	break;
      	case tR4:
				  OPC->AddFloat(id + stSysTagsCount, sl->Strings[1].c_str(), acc);
        	break;
      	case tR8:
				  OPC->AddDouble(id + stSysTagsCount, sl->Strings[1].c_str(), acc);
        	break;
      }
      sl->Clear();
  	}
  } __finally {
  	delete sl;
  }
}
//---------------------------------------------------------------------------
__fastcall TServer::~TServer()
{
  delete[] Types;
	delete OPC;
}
//---------------------------------------------------------------------------
//static
void __fastcall TServer::Configure()
{
	strcpy((char*)OPCClsID, Config->GetOPCClsID().c_str());
  if (LogFile->CheckFlag(lfServer))
		LogFile->Log("Server OPCClsID: " + AnsiString(OPCClsID));
	strcpy((char*)OPCProgID, Config->GetOPCProgID().c_str());
  if (LogFile->CheckFlag(lfServer))
		LogFile->Log("Server OPCProgID: " + AnsiString(OPCProgID));
  if (::UuidFromString(Config->GetOPCGUID().c_str(), (GUID*)&OPCGUID) !=
  		RPC_S_OK)
		throw Exception("Invalid OPCGUID: " + Config->GetOPCGUID());
  unsigned char *b;
  UuidToString((GUID*)&OPCGUID, &b);
  if (LogFile->CheckFlag(lfServer))
		LogFile->Log("Server OPCGUID: " + AnsiString((char*)(void*)b));
  RpcStringFree(&b);
}
//---------------------------------------------------------------------------
void __fastcall TServer::Start()
{
	OPC->Start();
}
//---------------------------------------------------------------------------
void __fastcall TServer::Stop()
{
}
//---------------------------------------------------------------------------
void __fastcall TServer::SetValues(TStrings *sl)
{
	int n = sl->Count;
  if (LogFile->CheckFlag(lfServer))
		LogFile->Log("TServer::SetValues() n=" + AnsiString(n));
  TStringList *sl1 = new TStringList();
  try {
  	try {
      for (int i = 0; i < n; ++i) {
        try {
				  if (LogFile->CheckFlag(lfServer))
						LogFile->Log("TServer::SetValues() val=" + sl->Strings[i]);
          if (::SplitString(sl->Strings[i], sl1, " ") != 5)
            throw Exception("TServer::SetValues() invalid value format: " +
                sl->Strings[i]);
          int id = sl1->Strings[0].ToIntDef(0);
          if (id < 1)
            throw Exception("TServer::SetValues() invalid value format: " +
                sl->Strings[i]);
          if (id > N)
            throw Exception("TServer::SetValues() invalid value index: " +
                sl->Strings[i]);
          TFileTime t;
          ::DateTimeToFileTime(sl1->Strings[3] + " " + sl1->Strings[4], t);
          bool gf;
          if (sl1->Strings[2] == "g")
          	gf = true;
          else if (sl1->Strings[2] == "b")
          	gf = false;
          else
            throw Exception("Invalid quality: " + sl1->Strings[1]);
          if (gf) {
            int v;
            switch(Types[id - 1]) {
              case tB:
                v = sl1->Strings[1].Pos(".") < 1 ?
                		sl1->Strings[1].ToInt() :
                		::Round(sl1->Strings[1].ToDouble());
                if (v < 0 || v > 1) {
                  OPC->SetBad(id + stSysTagsCount, &t);
                  throw Exception("Invalid bool value: " +
                      sl1->Strings[1]);
                } else
                  OPC->SetBool(id + stSysTagsCount, v, &t,
                      gf ? OPC_QUALITY_GOOD : OPC_QUALITY_BAD);
                break;
              case tI1:
                v = sl1->Strings[1].Pos(".") < 1 ?
                		sl1->Strings[1].ToInt() :
                		::Round(sl1->Strings[1].ToDouble());
                if (v < -128 || v > 127) {
                  OPC->SetBad(id + stSysTagsCount, &t);
                  throw Exception("Invalid byte value: " +
                      sl1->Strings[1]);
                } else
                  OPC->SetByte(id + stSysTagsCount, v, &t,
                      gf ? OPC_QUALITY_GOOD : OPC_QUALITY_BAD);
                break;
              case tU1:
                v = sl1->Strings[1].Pos(".") < 1 ?
                		sl1->Strings[1].ToInt() :
                		::Round(sl1->Strings[1].ToDouble());
                if (v < 0 || v > 255) {
                  OPC->SetBad(id + stSysTagsCount, &t);
                  throw Exception("Invalid unsigned byte value: " +
                      sl1->Strings[1]);
                } else
                  OPC->SetUByte(id + stSysTagsCount, v, &t,
                      gf ? OPC_QUALITY_GOOD : OPC_QUALITY_BAD);
                break;
              case tI2:
                v = sl1->Strings[1].Pos(".") < 1 ?
                		sl1->Strings[1].ToInt() :
                		::Round(sl1->Strings[1].ToDouble());
                if (v < -32768 || v > 32767) {
                  OPC->SetBad(id + stSysTagsCount, &t);
                  throw Exception("Invalid short value: " +
                      sl1->Strings[1]);
                } else
                  OPC->SetShort(id + stSysTagsCount, v, &t,
                      gf ? OPC_QUALITY_GOOD : OPC_QUALITY_BAD);
                break;
              case tU2:
                v = sl1->Strings[1].Pos(".") < 1 ?
                		sl1->Strings[1].ToInt() :
                		::Round(sl1->Strings[1].ToDouble());
                if (v < 0 || v > 65535) {
                  OPC->SetBad(id + stSysTagsCount, &t);
                  throw Exception("Invalid unsigned short value: " +
                      sl1->Strings[1]);
                } else
                  OPC->SetUShort(id + stSysTagsCount, v, &t,
                      gf ? OPC_QUALITY_GOOD : OPC_QUALITY_BAD);
                break;
              case tI4:
                v = sl1->Strings[1].Pos(".") < 1 ?
                		sl1->Strings[1].ToInt() :
                		::Round(sl1->Strings[1].ToDouble());
                OPC->SetLong(id + stSysTagsCount, v, &t,
                    gf ? OPC_QUALITY_GOOD : OPC_QUALITY_BAD);
                break;
              case tU4:
                v = sl1->Strings[1].Pos(".") < 1 ?
                		sl1->Strings[1].ToInt() :
                		::Round(sl1->Strings[1].ToDouble());
                OPC->SetULong(id + stSysTagsCount, v, &t,
                    gf ? OPC_QUALITY_GOOD : OPC_QUALITY_BAD);
                break;
              case tR4:
                OPC->SetFloat(id + stSysTagsCount, sl1->Strings[1].ToDouble(), &t,
                    gf ? OPC_QUALITY_GOOD : OPC_QUALITY_BAD);
                break;
              case tR8:
                OPC->SetDouble(id + stSysTagsCount, sl1->Strings[1].ToDouble(), &t,
                    gf ? OPC_QUALITY_GOOD : OPC_QUALITY_BAD);
                break;
            }
          } else
            OPC->SetBad(id + stSysTagsCount, &t);
        } catch(Exception &e) {
          LogFile->Log(e, "TServer::SetValues() value: " + sl->Strings[i]);
        }
        sl1->Clear();
      }
	  } __finally {
    	OPC->Flush();
	  }
  } __finally {
  	delete sl1;
  }
}
//---------------------------------------------------------------------------
void __fastcall TServer::SetBad()
{
  if (LogFile->CheckFlag(lfServer))
		LogFile->Log("TServer::SetBad()");
	try {
  	for (int i = 0; i < N; ++i)
      OPC->SetBad(i + stSysTagsCount + 1);
  } __finally {
   	OPC->Flush();
  }
}
//---------------------------------------------------------------------------
void __fastcall TServer::SetSTConnections(bool val)
{
  OPC->SetBool(stConnection, val);
 	OPC->Flush();
}
//---------------------------------------------------------------------------
void __fastcall TServer::SetSTTagsHashGood(bool val)
{
  OPC->SetBool(stTagsHashGood, val);
 	OPC->Flush();
}
//---------------------------------------------------------------------------

