//---------------------------------------------------------------------------
/*
Copyright (c) 2013 Serge L. Ryadkow admin@opcgate.ru http://opcgate.ru 

Permission is hereby granted, free of charge, to any person obtaining a copy of this 
software and associated documentation files (the "Software"), to deal in the Software 
without restriction, including without limitation the rights to use, copy, modify, 
merge, publish, distribute, sublicense, and/or sell copies of the Software, and to 
permit persons to whom the Software is furnished to do so, subject to the following 
conditions:

The above copyright notice and this permission notice shall be included in all copies 
or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, 
INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR 
PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE 
FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR 
OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR 
OTHER DEALINGS IN THE SOFTWARE. 
*/
//---------------------------------------------------------------------------
//---------------------------------------------------------------------------
#include <vcl.h>

#pragma hdrstop

#include <DateUtils.hpp>

#include "Protocol.h"
#include "Session.h"
#include "Gate.h"
//---------------------------------------------------------------------------
#pragma package(smart_init)
//---------------------------------------------------------------------------
namespace OPCGate {
//---------------------------------------------------------------------------
__fastcall TProtocol::TProtocol(TOPCGate *gate, int connTOms, int subsTOms,
		bool buffSubs)
	:	ConnTOms(connTOms),
		SubsTOms(subsTOms),
    BuffSubs(buffSubs)
{
	Gate = gate;
}
//---------------------------------------------------------------------------
__fastcall TProtocol::~TProtocol()
{
}
//---------------------------------------------------------------------------
bool __fastcall TProtocol::Execute(TIdPeerThread *thread)
{
  TSession *sess = dynamic_cast<TSession*>(thread->Data);
  if (!sess)
  	throw Exception("TProtocol::Execute() "
    		"dynamic_cast<TSession*>(thread->Data) == 0");
  bool subs;
  TDateTime pt;
  int cto = ConnTOms, sto = SubsTOms;
  sess->Lock();
  try {
    if ((subs = sess->Subscribed) == true)
      pt = sess->PingTime;
    if (sess->ConnTOms)
    	cto = sess->ConnTOms;
    if (sess->SubsTOms)
    	sto = sess->SubsTOms;
  } __finally {
    sess->Unlock();
  }
  AnsiString req = thread->Connection->ReadLn(EOL, subs ? sto : cto);
  if (req.IsEmpty()) {
    if (subs) {
    	if (MilliSecondsBetween(Now(), pt) > cto) {
        sess->Lock();
        try {
          if (BuffSubs)
	          sess->Values->Clear();
          sess->Subscribed = false;
        } __finally {
          sess->Unlock();
        }
      	return false;
      } else {
        if (BuffSubs) {
          sess->Lock();
          try {
            int n = sess->Values->Count;
            if (n) {
              thread->Connection->WriteLn("VALUES " + AnsiString(n));
              for (int i = 0; i < n; ++i)
                thread->Connection->WriteLn("VALUE " +
                		sess->Values->Strings[i]);
              sess->Values->Clear();
            }
          } __finally {
            sess->Unlock();
          }
        }
        return true;
      }
    } else
	  	return false;
  } else {
    int sp = req.Pos(" ");
    AnsiString cmd = sp > 0 ? req.SubString(1, sp - 1) : req;
    if (cmd == "QUIT") {
      HandleQuit(sess, thread->Connection);
      return false;
    } else {
      if (cmd == "PING")
        HandlePing(sess);
      else if (cmd == "HASH")
        HandleHash(sess, thread->Connection);
      else if (cmd == "TAGS")
        HandleTags(sess, thread->Connection);
      else if (cmd == "VALUES")
        HandleValues(sess, thread->Connection);
      else if (cmd == "SUBSCRIBE")
        HandleSubscr(sess, thread->Connection);
      else if (cmd == "UNSUBSCRIBE")
        HandleUnsubscr(sess, thread->Connection);
      else if (cmd == "PAR_CONNTO" && sp > 0)
        HandleParConnTO(sess, thread->Connection,
        		req.SubString(12, req.Length() - 11).ToInt());
      else if (cmd == "PAR_SUBSTO" && sp > 0)
        HandleParSubsTO(sess, thread->Connection,
        		req.SubString(12, req.Length() - 11).ToInt());
      return true;
    }
  }
}
//---------------------------------------------------------------------------
void __fastcall TProtocol::SendToSubscr(TIdPeerThread *thread, TStrings *sl)
{
  TSession *sess = dynamic_cast<TSession*>(thread->Data);
  if (!sess)
  	throw Exception("TProtocol::SendToSubscr() "
    		"dynamic_cast<TSession*>(thread->Data) == 0");
  int n = sl->Count;
  sess->Lock();
  try {
  	if (sess->Subscribed) {
      if (BuffSubs) {
        int nn = sess->Values->Count;
        for (int i = 0; i < n; ++i) {
          int id = sl->Strings[i].SubString(1,
              sl->Strings[i].Pos(" ") - 1).ToInt();
          bool f = false;
          for (int j = 0; j < nn && !f; ++j)
            if (reinterpret_cast<int>(sess->Values->Objects[j]) == id) {
              sess->Values->Strings[j] = sl->Strings[i];
              f = true;
            }
          if (!f)
            sess->Values->AddObject(sl->Strings[i],
                reinterpret_cast<TObject*>(id));
        }
      } else {
	      thread->Connection->WriteLn("VALUES " + AnsiString(n));
  	    for (int i = 0; i < n; ++i)
    	    thread->Connection->WriteLn("VALUE " + sl->Strings[i]);
      }
    }
  } __finally {
    sess->Unlock();
  }
}
//---------------------------------------------------------------------------
void __fastcall TProtocol::HandleQuit(TSession *sess,
		TIdTCPServerConnection *conn)
{
	sess->Lock();
  try {
    if (BuffSubs)
      sess->Values->Clear();
  	conn->WriteLn("QUIT_ACK");
  } __finally {
  	sess->Unlock();
  }
}
//---------------------------------------------------------------------------
void __fastcall TProtocol::HandlePing(TSession *sess)
{
  TDateTime now = Now();
	sess->Lock();
  try {
    sess->PingTime = now;
  } __finally {
  	sess->Unlock();
  }
}
//---------------------------------------------------------------------------
void __fastcall TProtocol::HandleHash(TSession *sess,
		TIdTCPServerConnection *conn)
{
	sess->Lock();
  try {
  	conn->WriteLn("HASH " + Gate->GetTagsHash());
  } __finally {
  	sess->Unlock();
  }
}
//---------------------------------------------------------------------------
void __fastcall TProtocol::HandleTags(TSession *sess,
		TIdTCPServerConnection *conn)
{
 	TStrings *sl = Gate->GetTags();
  int n = sl->Count;
	sess->Lock();
  try {
  	conn->WriteLn("TAGS " + AnsiString(n));
    for (int i = 0; i < n; ++i)
	  	conn->WriteLn("TAG " + sl->Strings[i]);
  } __finally {
  	sess->Unlock();
  }
}
//---------------------------------------------------------------------------
void __fastcall TProtocol::HandleValues(TSession *sess,
		TIdTCPServerConnection *conn)
{
	TStringList *sl = new TStringList();
  try {
  	Gate->GetValues(sl);
    int n = sl->Count;
    sess->Lock();
    try {
      conn->WriteLn("VALUES " + AnsiString(n));
      for (int i = 0; i < n; ++i)
        conn->WriteLn("VALUE " + sl->Strings[i]);
    } __finally {
      sess->Unlock();
    }
  } __finally {
  	delete sl;
  }
}
//---------------------------------------------------------------------------
void __fastcall TProtocol::HandleSubscr(TSession *sess,
		TIdTCPServerConnection *conn)
{
  TDateTime now = Now();
	sess->Lock();
  try {
		sess->Subscribed = true;
    sess->PingTime = now;
		if (BuffSubs)
	  	sess->Values->Clear();
  	conn->WriteLn("SUBSCRIBE_ACK");
  } __finally {
  	sess->Unlock();
  }
}
//---------------------------------------------------------------------------
void __fastcall TProtocol::HandleUnsubscr(TSession *sess,
		TIdTCPServerConnection *conn)
{
	sess->Lock();
  try {
		sess->Subscribed = false;
		if (BuffSubs)
	  	sess->Values->Clear();
  	conn->WriteLn("UNSUBSCRIBE_ACK");
  } __finally {
  	sess->Unlock();
  }
}
//---------------------------------------------------------------------------
void __fastcall TProtocol::HandleParConnTO(TSession *sess,
		TIdTCPServerConnection *conn, int v)
{
	sess->Lock();
  try {
    if (v <= ConnTOms) {
			sess->ConnTOms = v;
  		conn->WriteLn("PAR_CONNTO_ACK");
    } else
  		conn->WriteLn("PAR_CONNTO_ERR");
  } __finally {
  	sess->Unlock();
  }
}
//---------------------------------------------------------------------------
void __fastcall TProtocol::HandleParSubsTO(TSession *sess,
		TIdTCPServerConnection *conn, int v)
{
	sess->Lock();
  try {
    if (v <= SubsTOms) {
			sess->SubsTOms = v;
  		conn->WriteLn("PAR_SUBSTO_ACK");
    } else
  		conn->WriteLn("PAR_SUBSTO_ERR");
  } __finally {
  	sess->Unlock();
  }
}
//---------------------------------------------------------------------------
} //namespace OPCGate
//---------------------------------------------------------------------------
