//---------------------------------------------------------------------------
/*
Copyright (c) 2013 Serge L. Ryadkow admin@opcgate.ru http://opcgate.ru 

Permission is hereby granted, free of charge, to any person obtaining a copy of this 
software and associated documentation files (the "Software"), to deal in the Software 
without restriction, including without limitation the rights to use, copy, modify, 
merge, publish, distribute, sublicense, and/or sell copies of the Software, and to 
permit persons to whom the Software is furnished to do so, subject to the following 
conditions:

The above copyright notice and this permission notice shall be included in all copies 
or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, 
INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR 
PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE 
FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR 
OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR 
OTHER DEALINGS IN THE SOFTWARE. 
*/
//---------------------------------------------------------------------------
//---------------------------------------------------------------------------
#include <vcl.h>

#pragma hdrstop

#include <Registry.hpp>

#include <OPCClient.h>
#include <OPCEnum.h>

#include "MainU.h"
//---------------------------------------------------------------------------
#pragma package(smart_init)
#pragma resource "*.dfm"
TMainF *MainF;
//---------------------------------------------------------------------------
const AnsiString RegKey = "\\Software\\OPCGate\\OPCView";
//---------------------------------------------------------------------------
__fastcall TMainF::TMainF(TComponent* Owner)
	: TForm(Owner)
{
	RF = false;
  Application->OnIdle = OnIdle;
  AnsiString s;
  if (__argc > 1)
    s = __argv[1];
  try {
    TRegistry *r = new TRegistry;
    try {
      r->RootKey = HKEY_CURRENT_USER;
      if (r->OpenKeyReadOnly(RegKey)) {
        if (s.IsEmpty())
          s = r->ReadString("Server");
        ExpandCB->Checked = r->ReadBool("Expand");
      }
    }
    __finally {
      delete r;
    }
  }
  catch(...) {}
  TOPCEnum e;
  e.EnumProgId(ServerCB->Items, TOPCEnum::OPC_DA);
  int i = ServerCB->Items->IndexOf(s);
  if (i >= 0) {
  	ServerCB->ItemIndex = i;
    if (__argc > 1)
      StartBClick(StartB);
  }
}
//---------------------------------------------------------------------------
__fastcall TMainF::~TMainF()
{
	if (RF) {
		Timer1->Enabled = false;
		try {
  		Disconnect();
	  } catch(Exception &e) {
  	}
  }
  TRegistry *r = new TRegistry;
  try {
    r->RootKey = HKEY_CURRENT_USER;
    if (r->OpenKey(RegKey, true)) {
      r->WriteString("Server", ServerCB->Text);
      r->WriteBool("Expand", ExpandCB->Checked);
    }
  }
  __finally {
    delete r;
  }
}
//---------------------------------------------------------------------------
void __fastcall TMainF::Connect()
{
  UF = false;
	Nodes = 0;
  NodeCnt = 0;
  Group = 0;
	Server = new TOPCServer(ServerCB->Text);
  TOPCBrowser browser(Server);
  TTreeNode *root = Tree->Items->Add(0, ServerCB->Text);
  browser.FillTreeObj<TNodeData>(Tree->Items, root);
	if (ExpandCB->Checked)
	  Tree->FullExpand();
  TStringList *sl = new TStringList();
  try {
    browser.FillList(sl);
		NodeCnt = sl->Count;
    Nodes = new PTreeNode[NodeCnt];
    Group = new TAsyncOPCGroup(Server, sl, "Group1");
    Group->OnDataChange = OnChange;
  } __finally {
  	delete sl;
  }
  int l = 0;
  for (int i = 0, j = 0, n = Tree->Items->Count; i < n; ++i) {
    TTreeNode *n = Tree->Items->Item[i];
    n->StateIndex = -1;
  	if (Tree->Items->Item[i]->Data) {
      Nodes[j] = n;
      static_cast<TNodeData*>(n->Data)->Tag = n->Text;
      int l1 = n->Text.Length();
      if (l1 > l)
      	l = l1;
      switch (Group->Value[j].Type()) {
       	case varEmpty:
		      n->ImageIndex = 0;
		      n->SelectedIndex = 0;
          break;
        case varBoolean:
        	n->ImageIndex = 2;
		      n->SelectedIndex = 2;
          break;
        default:
        	n->ImageIndex = 3;
		      n->SelectedIndex = 3;
      }
      ++j;
    } else {
    	if (n == root) {
	    	n->ImageIndex = 4;
  	    n->SelectedIndex = 4;
      } else {
	    	n->ImageIndex = 1;
  	    n->SelectedIndex = 1;
      }
    }
  }
  for (int i = 0; i < NodeCnt; ++i) {
		TNodeData *d = static_cast<TNodeData*>(Nodes[i]->Data);
    int l1 = d->Tag.Length();
    if (l1 < l)
    	d->Tag += AnsiString::StringOfChar(' ', l - l1);
  }
	ServerTimeST->Caption = Server->CurrentTime.FormatString(
  		"dd\".\"mm\".\"yyyy\" \"hh\":\"nn\":\"ss");
}
//---------------------------------------------------------------------------
void __fastcall TMainF::Disconnect()
{
	if (NodeCnt)
  	delete Nodes;
  delete Group;
	delete Server;
  Server = 0;
  Tree->Items->Clear();
	ServerTimeST->Caption = "?";
}
//---------------------------------------------------------------------------
void __fastcall TMainF::OnChange(TAsyncOPCGroup *group, int count, int *i)
{
  UF = true;
  for (int j = 0; j < count; ++j) {
    Nodes[i[j]]->Text = static_cast<TNodeData*>(Nodes[i[j]]->Data)->Tag +
        "   " +
        group->Time[i[j]].FormatString(
            "dd\".\"mm\".\"yyyy\" \"hh\":\"nn\":\"ss\".\"zzz") +
        (group->Good[i[j]] ? "   GOOD   " : "   BAD    ") +
        AnsiString(group->Value[i[j]]);
    if (!Nodes[i[j]]->ImageIndex) {
      Nodes[i[j]]->ImageIndex =
          group->Value[i[j]].Type() == varBoolean ? 2 : 3;
      Nodes[i[j]]->SelectedIndex = Nodes[i[j]]->ImageIndex;
    }
  }
}
//---------------------------------------------------------------------------
void __fastcall TMainF::Timer1Timer(TObject *Sender)
{
	if (Server) {
  	if (!UF) {
	    try {
		    Group->Refresh();
    	} catch(...) {}
    }
    bool f = false;
  	try {
			ServerTimeST->Caption = Server->CurrentTime.FormatString(
  				"dd\".\"mm\".\"yyyy\" \"hh\":\"nn\":\"ss");
    } catch(Exception &e) {
    	f = true;
    }
    if (f) {
      try {
      	Disconnect();
      } catch(Exception &e) {
      }
    }
  } else {
    try {
     	Connect();
    } catch(Exception &e) {
    }
  }
}
//---------------------------------------------------------------------------
void __fastcall TMainF::Label3Click(TObject *Sender)
{
  ShellExecute(0, "open", "http://opcgate.ru", 0, 0, 0);
}
//---------------------------------------------------------------------------
void __fastcall TMainF::StartBClick(TObject *Sender)
{
	try {
  	Connect();
  } catch(Exception &e) {
  }
  Timer1->Enabled = true;
  RF = true;
  Caption = "OPCView " + ServerCB->Text;
}
//---------------------------------------------------------------------------
void __fastcall TMainF::StopBClick(TObject *Sender)
{
  Caption = "OPCView";
	RF = false;
	Timer1->Enabled = false;
	try {
  	Disconnect();
  } catch(Exception &e) {
  }
}
//---------------------------------------------------------------------------
void __fastcall TMainF::OnIdle(System::TObject* Sender, bool &Done)
{
	if (RF) {
  	if (!StopB->Enabled)
    	StopB->Enabled = true;
    if (StartB->Enabled)
    	StartB->Enabled = false;
  } else {
  	if (StopB->Enabled)
    	StopB->Enabled = false;
    if (ServerCB->Text.IsEmpty()) {
	    if (StartB->Enabled)
  	  	StartB->Enabled = false;
    } else {
	    if (!StartB->Enabled)
  	  	StartB->Enabled = true;
    }
  }
}
//---------------------------------------------------------------------------

